from typing import Any, List

from pydantic import BaseModel

from .StatusMessage import StatusMessage
from .utils import to_camel


class Failure(BaseModel):
    """
    Details of a failed task
    """

    errors: List[StatusMessage]
    """*(required)* List of status messages describing the failure"""

    class Config:
        alias_generator = to_camel

    def json(self, **kwargs: Any) -> str:
        if "exclude_none" not in kwargs.keys():
            kwargs["exclude_none"] = True
        return super().json(**kwargs)

    def __getitem__(self, key):
        return getattr(self, key)

    def get(self, key, value: Any = None):
        if hasattr(self, key):
            return getattr(self, key)
        return value



import sys
import traceback
from argparse import ArgumentParser
from typing import Any, List

from loguru import logger

from . import BaseELGCommand


def local_installation_command_factory(args):
    return LocalInstallationDockerCommand(
        image=args.image,
        execution_location=args.execution_location,
        sidecar_image=args.sidecar_image,
        image_envvars=args.image_envvars,
        name=args.name,
        full_name=args.full_name,
        gui=args.gui,
        gui_image=args.gui_image,
        gui_port=args.gui_port,
        gui_path=args.gui_path,
        folder=args.folder,
        expose_port=args.expose_port,
        helpers=args.helper,
    )


class LocalInstallationDockerCommand(BaseELGCommand):
    @staticmethod
    def register_subcommand(parser: ArgumentParser):
        local_installation_parser = parser.add_parser(
            "docker",
            description="Create a Docker compose file to deploy an ELG compatible Docker image locally",
        )
        local_installation_parser.add_argument(
            "--image",
            type=str,
            required=True,
            help="name of the Docker image",
        )
        local_installation_parser.add_argument(
            "--execution_location",
            type=str,
            required=True,
            help="endpoint of the Docker image where the service can be executed",
        )
        local_installation_parser.add_argument(
            "--image_envvars",
            type=str,
            default=[],
            required=False,
            nargs="*",
            help='environment variables to be set in the docker image. format "key=value"',
        )
        local_installation_parser.add_argument(
            "--sidecar_image",
            type=str,
            default="",
            required=False,
            help="name of the sidecare Docker image",
        )
        local_installation_parser.add_argument(
            "--name",
            type=str,
            default=None,
            required=False,
            help="name of the service to use for the REST server",
        )
        local_installation_parser.add_argument(
            "--full_name",
            type=str,
            default=None,
            required=False,
            help="name of the service to display in the GUI",
        )
        local_installation_parser.add_argument(
            "--folder",
            type=str,
            default="./elg_local_installation/",
            required=False,
            help="path to the folder where to save the Docker compose file",
        )
        local_installation_parser.add_argument(
            "--gui",
            default=False,
            action="store_true",
            required=False,
            help="use to use the GUI",
        )
        local_installation_parser.add_argument(
            "--expose_port",
            type=int,
            default=8080,
            required=False,
            help="port used to publish to the host",
        )
        local_installation_parser.add_argument(
            "--gui_image",
            type=str,
            default="registry.gitlab.com/european-language-grid/usfd/gui-ie:latest",
            required=False,
            help="docker image of the GUI",
        )
        local_installation_parser.add_argument(
            "--gui_path",
            type=str,
            default="",
            required=False,
            help="path to the GUI",
        )
        local_installation_parser.add_argument(
            "--gui_port",
            type=int,
            default=80,
            required=False,
            help="port used by the GUI docker image",
        )
        local_installation_parser.add_argument(
            "--helper",
            type=str,
            required=False,
            default=[],
            action="append",
            choices=["temp-storage"],
            help="helper services to include in the compose stack",
        )

        local_installation_parser.set_defaults(func=local_installation_command_factory)

    def __init__(
        self,
        image: str,
        execution_location: str,
        folder: str = "./",
        gui: bool = False,
        expose_port: int = 8080,
        sidecar_image: str = "",
        name: str = None,
        image_envvars: List[str] = [],
        full_name: str = None,
        gui_image: str = "registry.gitlab.com/european-language-grid/usfd/gui-ie:latest",
        gui_port: int = 80,
        gui_path: str = "",
        helpers: List[str] = None,
        record: Any = {},
    ):
        self._folder = folder
        self._gui = gui
        self._expose_port = expose_port
        self._image = image
        self._execution_location = execution_location
        self._sidecar_image = sidecar_image
        self._name = name
        self._image_envvars = image_envvars
        self._full_name = full_name
        self._gui_image = gui_image
        self._gui_port = gui_port
        self._gui_path = gui_path
        self._helpers = helpers
        self._record = record

    def run(self):
        from ..local_installation import (LocalInstallation,
                                          LTServiceLocalInstallation)

        try:
            ltservice = LTServiceLocalInstallation.from_docker_image(
                image=self._image,
                execution_location=self._execution_location,
                sidecar_image=self._sidecar_image,
                name=self._name,
                image_envvars=self._image_envvars,
                full_name=self._full_name,
                gui=self._gui,
                gui_image=self._gui_image,
                gui_port=self._gui_port,
                gui_path=self._gui_path,
                record=self._record,
            )
            LocalInstallation(ltservices=[ltservice], helpers=self._helpers).create_docker_compose(
                expose_port=self._expose_port,
                path=self._folder,
            )
        except Exception as e:
            logger.error(f"Error during the creation of the Docker compose file - {e}")
            traceback.print_exc()
            sys.exit(1)



from pathlib import Path
from typing import Any, Dict, List

from .. import Annotation, Request


class TextRequest(Request):
    """
    Request representing a single piece of text, optionally with associated markup
    Subclass of :class:`elg.model.base.Request.Request`

    For example a list of paragraphs or sentences, or a corpus of documents, each divided into sentences.
    While this could be represented as standoff annotations in a plain "text" request, the structured format is more
    suitable for certain types of tools.
    """

    type: str = "text"
    """*(required)* the type of request must be \"text\""""

    content: str
    """*(optional)* text content"""

    mimeType: str = "text/plain"
    """*(optional)* mime type of request, default \"text/plain\""""

    features: Dict = None
    """*(optional)* arbitrary json metadata about content"""

    annotations: Dict[str, List[Annotation]] = None
    """*(optional)* optional annotations on request"""

    @classmethod
    def from_file(
        cls,
        filename,
        features: Dict = None,
        annotations: Dict[str, List[Annotation]] = None,
        params: dict = None,
    ):
        filename = Path(filename)
        if not filename.is_file():
            raise ValueError(f"{filename} musts be the path to a file.")
        with open(filename) as f:
            content = f.read()
        return cls(content=content, features=features, annotations=annotations, params=params)

    @classmethod
    def from_auto_content(cls, content: Any):
        if isinstance(content, str):
            return cls(content=content)
        elif isinstance(content, dict):
            return cls(**content)
        else:
            raise ValueError(f"The TextRequest message cannot be initialize from {content}")



import json

import requests
from jproperties import Properties

"""
Fairly basic script to generate elg/model/base/StandardMessages from the standard messages.properties file in `elg-apis`.
Left simple to make maintenance/editing easy.

To run from the root directory of this project: `python elg/utils/_generate_standard_messages.py`.

Requires modules "requests" (to download the properties file) and "jproperties" (to parse it).
"""
properties_content = requests.get(
    "https://gitlab.com/european-language-grid/platform/elg-apis/-/raw/master/java/src/main/resources/eu/elg/elg-messages.properties"
).content

props = Properties()
props.load(properties_content, "utf-8")

details = {k: v.replace("''", "'") for (k, (v, _)) in props.items()}


def stringliteral(v):
    """Prefer double quoted string (json style) unless the string contains any double quotes."""
    if '"' in str(v):
        return repr(v)
    else:
        return json.dumps(v)


with open("elg/model/base/StandardMessages.py", "w+") as file:
    file.write("from .StatusMessage import StatusMessage\n\n")
    file.write("\nclass StandardMessages:\n")
    file.write(
        '\n    """\n'
        + "    This class provides easy access to the standard set of ELG status messages that are provided by default by\n"
        + "    the platform and should be fully translated in the ELG user interface. If you use codes other than these\n"
        + "    standard ones in your services then you should also try to contribute translations of your messages into as\n"
        + "    many languages as possible for the benefit of other ELG users.\n\n"
        + "    Implementation note: This class is auto-generated from elg-messages.properties - to add new message codes you\n"
        + "    should edit the property files, then run /utils/generate_standard_messages.py. Do not edit this class directly.\n"
        + '    """\n'
    )
    for key in details:
        file.write("\n    @classmethod")
        file.write("\n    def generate_" + key.replace(".", "_").lower() + "(cls, params=[], detail={}, **kwargs):")
        file.write('\n        """Generate StatusMessage for code: ' + key + '"""')
        file.write('\n        if "lang" in kwargs:')
        file.write("\n            print(")
        file.write(
            "\n                \"Warning: lang is deprecated and will be ignored. The standard message 'text' will be in English, pass it to the i18n resolver to obtain it in another language.\""
        )
        file.write("\n            )")
        file.write("\n        code = " + stringliteral(key))
        file.write("\n        text = " + stringliteral(details[key]))
        file.write("\n        return StatusMessage(code=code, text=text, params=params, detail=detail)")
        file.write("\n")



from typing import Any, Dict, List

from pydantic import BaseModel

from .utils import to_camel


class StatusMessage(BaseModel):
    """
    Represents a single status message, in a form amenable to internationalisation.

    Each message contains a code, which can be looked up in a list to find the actual text in any of the available
    languages.  The text can contain numbered placeholders of the form <code>{0}</code>, which are filled in with values
    specified in the "params" property. The "text" property provides a single fallback text to be used if the specified
    code cannot be found in the lookup table.
    """

    code: str
    """*(required)* status code to be found in lookup table"""

    params: List[str]
    """*(required)* values to fill in message placeholder"""

    text: str
    """*(required)* fallback text to be used if specified code cannot be found in lookup table"""

    detail: Dict = None
    """*(optional)* arbitrary further details that don't need translation (e.g. stacktrace)"""

    class Config:
        alias_generator = to_camel

    def json(self, **kwargs: Any) -> str:
        if "exclude_none" not in kwargs.keys():
            kwargs["exclude_none"] = True
        return super().json(**kwargs)

    def __getitem__(self, key):
        return getattr(self, key)

    def get(self, key, value: Any = None):
        if hasattr(self, key):
            return getattr(self, key)
        return value



from typing import Any, Dict, List

from pydantic import BaseModel, root_validator

from .. import Annotation, Request
from ..base.utils import to_camel


class Text(BaseModel):
    """
    A single node in a structured text request.

    Each text can have an associated score (for example a confidence value for multiple alternative translations or
    transcriptions) and optional annotations, which can be linked to both the result text in this object and to the
    original source material from the corresponding request.
    """

    content: str = None
    """*(optional)* text content"""

    mime_type: str = "text/plain"
    """*(optional)* mime type of request, default \"text/plain\""""

    features: dict = None
    """*(optional)* arbitrary json metadata about content"""

    annotations: Dict[str, List[Annotation]] = None
    """*(optional)* optional annotations on request"""

    texts: "List[Text]" = None
    """*(optional)* recursive, same structure"""

    class Config:
        alias_generator = to_camel

    @root_validator()
    def either_content_or_text(cls, values):
        """
        ensures only either the "content" or the "text" fields are present
        """
        content, texts = values.get("content"), values.get("texts")
        if content is None and texts is None:
            raise ValueError('A structured text request must have either "content" or "texts" fields')
        if content:
            values["mimeType"] = "text/plain"
        else:
            values["mimeType"] = None
        return values

    def json(self, **kwargs: Any) -> str:
        if "exclude_none" not in kwargs.keys():
            kwargs["exclude_none"] = True
        return super().json(**kwargs)

    def __getitem__(self, key):
        return getattr(self, key)

    def get(self, key, value: Any = None):
        if hasattr(self, key):
            return getattr(self, key)
        return value


Text.update_forward_refs()


class StructuredTextRequest(Request):
    """
    Request representing text with some structure.
    Subclass of :class:`elg.model.base.Request.Request`

    For example a list of paragraphs or sentences, or a corpus of documents, each divided into sentences.
    Whilst this could be represented as standoff annotations in a plain "text" request, the structured format is more
    suitable for certain types of tools.
    """

    type: str = "structuredText"
    """*(required)* the type of request must be \"structuredText\""""

    texts: List[Text]
    """*(required)* the actual text object with the text content"""

    @classmethod
    def from_auto_content(cls, content: Any):
        if isinstance(content, list):
            return cls(texts=content)
        elif isinstance(content, dict):
            return cls(texts=[Text(**content)])
        else:
            raise ValueError(f"The StructuredTextRequest message cannot be initialize from {content}")



DOCKER_COMPOSE = """\
version: "3.7"

services:
  
{LTSERVICES}

{LTSERVICES_SIDECAR}

  restserver:
    image: registry.gitlab.com/european-language-grid/ilsp/elg-lt-service-execution-all:production-reactive
    command:
    - "--spring.webflux.base-path=/execution"
    - "--logging.level.elg.ltserviceexecution.api=WARN"
{LTSERVICES_URL}
    - "--elg.base.url=http://localhost:{EXPOSE_PORT}/execution"
{EXPOSE_PORT_CONFIG}
    restart: always

{GUIS}

{HELPERS}

{FRONTEND}

"""

EXPOSE_PORT_CONFIG = """\
    ports:
    - "{EXPOSE_PORT}:8080"
"""

LTSERVICE = """\
  {LTSERVICE_NAME}:
    image: "{LTSERVICE_IMAGE}"
    environment: {LTSERVICE_ENVVARS}
    restart: always\
"""

LTSERVICE_SIDECAR = """\
  {LTSERVICE_NAME}_sidecar:
    image: "{SIDECAR_IMAGE}"
    network_mode: "service:{LTSERVICE_NAME}"
    depends_on:
      - {LTSERVICE_NAME}
"""

LTSERVICE_URL = (
    '    - "--elg.ltservices.staticServices.{LTSERVICE_NAME}=http://{LTSERVICE_NAME}:{LTSERVICE_PORT}{LTSERVICE_PATH}"'
)

HELPERS = {
    "temp-storage": """\
  tempstore:
    image: registry.gitlab.com/european-language-grid/platform/temporary-storage:latest
    command:
    - "--base-dir=/mnt"
    - "--upload-address=:80"
    - "--url-base=http://localhost:{EXPOSE_PORT}/storage/retrieve/"
    - "--cleanup-interval=10m"
    restart: always
    networks:
      default:
        aliases:
        - "storage.elg"
""",
}

GUI = """\
  {GUI_NAME}:
    image: "{GUI_IMAGE}"
    restart: always
"""

I18N = """\
  i18n:
    image: registry.gitlab.com/european-language-grid/platform/i18n-service:latest
    restart: always
"""

FRONTEND = """\
  frontend:
    image: "nginx:alpine"
    ports:
    - "{EXPOSE_PORT}:80"
    volumes:
    - type: bind
      source: ./nginx-conf
      target: /etc/nginx/templates
    restart: always
"""

DEFAULT_CONF_TEMPLATE = """\
log_format upstream_logging '[$time_local] $remote_addr - $remote_user - $server_name to: $upstream_addr: $request upstream_response_time $upstream_response_time msec $msec request_time $request_time';

server {{
    listen       80;
    server_name  localhost;

    location / {{
        alias /etc/nginx/conf.d/html/;
        index  index.html index.htm;
    }}

    location ~ \.json {{
        root /etc/nginx/templates/records;
    }}

    location /execution/ {{
        access_log /dev/stdout upstream_logging;
        proxy_pass http://restserver:8080/execution/;
    }}

{GUIS}

{HELPERS}

    # redirect server error pages to the static page /50x.html
    #
    error_page   500 502 503 504  /50x.html;
    location = /50x.html {{
        root   /usr/share/nginx/html;
    }}
}}
"""

GUI_CONF_TEMPLATE = """\
    location /{GUI_NAME}/ {{
          access_log /dev/stdout upstream_logging;
          proxy_pass http://{GUI_NAME}:{GUI_PORT}/;
      }}
"""

I18N_CONF_TEMPLATE = """\
    location /i18n/ {
        proxy_pass http://i18n:8080/;
    }
"""

HELPERS_TEMPLATE = {
    "temp-storage": """\
    location /storage/ {
        access_log /dev/stdout upstream_logging;
        proxy_pass http://tempstore:8081/;
    }
""",
}

HTML_INDEX_HTML_TEMPLATE = """\
<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <title>ELG Local Installation</title>
    <link rel="preconnect" href="https://fonts.googleapis.com">
    <link rel="preconnect" href="https://fonts.gstatic.com" crossorigin>
    <link href="https://fonts.googleapis.com/css2?family=Noto+Sans+Display:wght@600&display=swap" rel="stylesheet">
    <style>
      body {{
        font-family: 'Noto Sans Display', sans-serif;
        background-color: #f8f4ed;
        display: flex;
        flex-direction: column;
      }}
      body > * {{
        margin-inline: auto;
      }}
      h1,h2 {{
        font-weight: 600;
        width: 900px;
        text-align: center;
      }}
      h1 {{
        font-size: 32px;
        margin-bottom: 32px;
      }}
      h2 {{
        font-size: 24px;
      }}
      iframe {{
        margin-bottom: 32px;
        background-color: white;
      }}
    </style>
    
</head>
<body>
<h1>European Language Grid<br/>Local Installation</h1>
{IFRAMES}
<script>
{SCRIPTS}
</script>
</body>
</html>
"""

HTML_INDEX_IFRAME = """\
<h2>[{LTSERVICE_ID}] {LTSERVICE_FULL_NAME}</h2>
<iframe width="900" height="500" id="{LTSERVICE_NAME}" src="about:blank"></iframe>
"""

HTML_INDEX_SCRIPT = """\
(function() {{
  var iframe = document.getElementById('{LTSERVICE_NAME}');
  window.addEventListener('message', function (e) {{
    if(iframe && iframe.contentWindow && e.source === iframe.contentWindow) {{
      // this is a configuration request from {LTSERVICE_NAME}
      iframe.contentWindow.postMessage(JSON.stringify({{
        "StyleCss":" ",
        "ServiceUrl":window.location.origin+"/execution/async/process/{LTSERVICE_NAME}",
        "ApiRecordUrl":window.location.origin+"/{LTSERVICE_NAME}.json"
      }}), window.location.origin);
    }}
  }}, false);

  iframe.src = '/{GUI_NAME}/{GUI_PATH}';
}})();
"""



from numbers import Number
from typing import Any

from pydantic import BaseModel

from .utils import to_camel


class Annotation(BaseModel):
    """
    Representation of a single annotation with respect to either one or two streams of data.

    Attributes
    ----------
    start (Number, required): annotation start location (in response)
    end (Number, required): annotation end location (in response)
    source_start (Number, required in cases): annotation start location (in source)
    source_end (Number, required in cases): annotation end location (in source)
    features (dict, optional): arbitrary json metadata about content
    """

    start: Number
    end: Number
    source_start: Number = None
    source_end: Number = None
    features: dict = None

    class Config:
        alias_generator = to_camel
        arbitrary_types_allowed = True

    def json(self, **kwargs: Any) -> str:
        if "exclude_none" not in kwargs.keys():
            kwargs["exclude_none"] = True
        return super().json(**kwargs)

    def __getitem__(self, key):
        return getattr(self, key)

    def get(self, key, value: Any = None):
        if hasattr(self, key):
            return getattr(self, key)
        return value



import inspect
import json
import sys
from collections.abc import AsyncIterable
from pathlib import Path
from typing import Dict, List

from loguru import logger

try:
    import docker
    from quart import Quart, current_app, g, make_response
    from quart import request as input_request
    from requests_toolbelt import MultipartDecoder
    from werkzeug.exceptions import BadRequest, RequestEntityTooLarge
    from werkzeug.http import parse_options_header
    from werkzeug.sansio.multipart import (Data, Epilogue, Field, File,
                                           MultipartDecoder, NeedData)
except:
    raise ImportError(
        "Extra dependencies need to be install to use the QuartService class. Please run: `pip install elg[quart]`."
    )

from .model import (AudioRequest, Failure, ImageRequest, Progress,
                    ResponseObject, StandardMessages, StructuredTextRequest,
                    TextRequest)
from .utils.docker import COPY_FOLDER, DOCKERFILE, ENTRYPOINT_QUART, ENV_QUART
from .utils.json_encoder import json_encoder


class ProcessingError(Exception):
    def __init__(self, status_code, standard_message):
        self.status_code = status_code
        self.standard_message = standard_message

    @staticmethod
    def InternalError(text, detail={}):
        return ProcessingError(
            500,
            StandardMessages.generate_elg_service_internalerror(params=[text], detail=detail),
        )

    @staticmethod
    def InvalidRequest(detail={}):
        return ProcessingError(400, StandardMessages.generate_elg_request_invalid(detail=detail))

    @staticmethod
    def TooLarge(detail={}):
        return ProcessingError(413, StandardMessages.generate_elg_request_too_large(detail=detail))

    @staticmethod
    def UnsupportedMime(mime, detail={}):
        return ProcessingError(
            400,
            StandardMessages.generate_elg_request_text_mimetype_unsupported(params=[mime], detail=detail),
        )

    @staticmethod
    def UnsupportedType(request_type, detail={}):
        return ProcessingError(
            400,
            StandardMessages.generate_elg_request_type_unsupported(params=[request_type], detail=detail),
        )

    def to_json(self):
        return {"failure": Failure(errors=[self.standard_message])}


class QuartService:
    """
    Class to help the creation of an ELG compatible service from a python model using Quart.
    Extra dependencies need to be install to use the QuartService class. Please run: `pip install elg[quart]`.

    The QuartService class is suitable for services that execute the request directly, for example of a simple language detection service::

        from elg import QuartService
        from elg.model import AnnotationsResponse
        import langdetect

        class ELGService(QuartService):
            async def process_text(self, content):
                langs = langdetect.detect_langs(content.content)
                ld = {}
                for l in langs:
                    ld[l.lang] = l.prob
                return AnnotationsResponse(features=ld)

        service = ELGService("LangDetection")
        app = service.app

    The QuartService class is also particulary useful for proxy services that forward the request to the actual LT service.
    For example a proxy for a Speech-to-text service running outside the ELG cluster::

        import traceback
        import aiohttp

        from loguru import logger

        from elg import QuartService
        from elg.model import TextsResponse
        from elg.quart_service import ProcessingError


        class Proxy(QuartService):

            consume_generator = False

            async def setup(self):
                self.session = aiohttp.ClientSession()

            async def shutdown(self):
                if self.session is not None:
                    await self.session.close()

            async def process_audio(self, content):
                try:
                    # Make the remote call
                    async with self.session.post("https://example.com/endpoint", data=content.generator) as client_response:
                        status_code = client_response.status
                        content = await client_response.json()
                except:
                    traceback.print_exc()
                    raise ProcessingError.InternalError('Error calling API')

                if status_code >= 400:
                    # if your API returns sensible error messages you could include that
                    # instead of the generic message
                    raise ProcessingError.InternalError('Error calling API')

                logger.info("Return the text response")
                return TextsResponse(texts=[{"content": content["text"]}])

        service = Proxy("Proxy")
        app = service.app

    """

    requirements = ["elg[quart]"]
    consume_generator = True

    def __init__(
        self,
        name: str = "My ELG Service",
        path: str = "/process",
        request_size_limit: int = None,
    ):
        """
        Init function of the QuartService

        Args:
            name (str, optional): Name of the service. It doesn't have any importance.
            path (str, optional): Path of the endpoint to expose. It can contain parameters, e.g., "/translate/<src>/<target>"
                later accessible in the process_* methods using the `self.url_param` method, e.g.,
                self.url_param('src'). Default to "/process".
        """
        self.name = name
        self.app = Quart(name)
        # Don't add an extra "status" property to JSON responses - this would break the API contract
        self.app.config["JSON_ADD_STATUS"] = False
        # Don't sort properties alphabetically in response JSON
        self.app.config["JSON_SORT_KEYS"] = False
        self.app.json_encoder = json_encoder(self)

        if request_size_limit is not None:
            self.app.config["MAX_CONTENT_LENGTH"] = request_size_limit

        # Exceptions handling
        self.app.register_error_handler(ProcessingError, self.error_message)
        self.app.register_error_handler(BadRequest, lambda err: self.error_message(ProcessingError.InvalidRequest()))
        self.app.register_error_handler(
            RequestEntityTooLarge,
            lambda err: self.error_message(ProcessingError.TooLarge()),
        )

        self.app.before_serving(self.setup)
        self.app.after_serving(self.shutdown)

        self.app.add_url_rule("/health", "health", self.health, methods=["GET"])
        self.app.add_url_rule(path, "process", self.process, methods=["POST"])

    def to_json(self, obj):
        """Hook that can be overridden by subclasses to customise JSON encoding.

        FlaskService can convert the following types to JSON by default, in
        addition to the types handled natively by `json.dump`:

        - date, time, datetime (via `.isoformat()`)
        - uuid.UUID (converted to `str`)
        - any pydantic.BaseModel including the ELG message types (via
          `.dict(by_alias=True, exclude_none=True)`)
        - anything Iterable (as a list)
        - any `dataclass` (converted to a `dict` via `dataclasses.asdict`)
        - anything with a `__json__` or `for_json` method (which is expected to
          return a serializable type)

        To handle other types, or to change the standard behaviour for any of
        the above types, subclasses can override this method, which will be
        called whenever an object other than a string, number, bool, list or
        dict must be serialized and is expected to return a JSON-serializable
        object to be used in place of the original, or `None` to fall back to
        the default behaviour.

        The default implementation of this method always returns `None`.

        Args:
            obj: the object to convert

        Returns:
            a substitute object suitable for JSON serialization, or `None` to
            use the default behaviour.
        """
        return None

    def run(self):
        """
        Method to start the app.
        """
        self.app.run()

    @staticmethod
    def error_message(err):
        logger.error("Return error message: {err}", err=err)
        return err.to_json(), err.status_code

    async def setup(self):
        """
        One-time setup tasks that must happen before the first request is
        handled. For example, it is possible to open an `aiohttp` `ClientSessions()`
        to use it :
        `self.session = aiohttp.ClientSession()`
        """
        pass

    async def shutdown(self):
        """
        Logic that must run at shutdown time, after the last request has been
        handled. For example closing the `aiohttp` `ClientSessions()`:
        ```
        if self.session is not None:
            await self.session.close()
        ```
        """
        pass

    async def convert_async_generator_to_bytes(self, generator):
        data = b""
        async for value in generator:
            data += value
        return data

    def health(self):
        return {"alive": True}

    @staticmethod
    async def parse_plain(audioformat):
        return {"type": "audio", "format": audioformat}, input_request.body

    @staticmethod
    async def parse_plain_image(image_format):
        return {"type": "image", "format": image_format}, input_request.body

    @staticmethod
    async def parse_multipart():
        boundary = input_request.mimetype_params.get("boundary", "").encode("ascii")
        if not boundary:
            raise ProcessingError.InvalidRequest()

        async def multipart_gen():
            """
            Logic: the request should consist of first a "form field" part named
            "request" containing JSON, and second a "file upload" part named
            "content" containing the audio or image. This generator fully parses the JSON
            part and yields that as a dict, then subsequently yields chunks of the
            audio/image data until they run out.  We create the generator and consume its
            first yield (the parsed JSON), then return the active generator so the
            rest of the binary chunks can be consumed by the caller in an async for.
            This logic is heavily inspired by quart.formparser but we don't use that
            directly as it would attempt to buffer the binary data rather than
            allowing it to stream directly from the request.
            """
            parser = MultipartDecoder(boundary, current_app.config["MAX_CONTENT_LENGTH"])

            found_request = False
            request_buf = []
            in_content = False
            async for data in input_request.body:
                parser.receive_data(data)
                event = parser.next_event()
                while not isinstance(event, (Epilogue, NeedData)):
                    if isinstance(event, Field):
                        # this should be the "request" section
                        if event.name != "request":
                            raise ProcessingError.InvalidRequest()
                        found_request = True
                        request_charset = "utf-8"
                        request_content_type = event.headers.get("content-type")
                        if request_content_type:
                            mimetype, ct_params = parse_options_header(request_content_type)
                            if mimetype != "application/json":
                                raise ProcessingError.InvalidRequest()
                            request_charset = ct_params.get("charset", request_charset)
                    elif isinstance(event, File):
                        if not found_request:
                            raise ProcessingError.InvalidRequest()
                        # this should be the "content" section
                        if event.name != "content":
                            raise ProcessingError.InvalidRequest()
                        in_content = True
                    elif isinstance(event, Data):
                        if in_content:
                            # we're streaming the content now
                            yield event.data
                        elif found_request:
                            request_buf.append(event.data)
                            if not event.more_data:
                                # finished the request section, so parse it
                                try:
                                    yield json.loads(b"".join(request_buf).decode(request_charset))
                                except:
                                    raise ProcessingError.InvalidRequest()
                                # allow the JSON buffer to be freed by the GC
                                request_buf = []
                    event = parser.next_event()

        mp_gen = multipart_gen()

        # generator yields first the parsed JSON request, then the content as
        # chunks of bytes
        return (await mp_gen.asend(None)), mp_gen

    def url_param(self, name: str):
        """
        Method to get give access to url parameters
        """
        return g._elg_args[name]

    async def process(self, **kwargs):
        """
        Main request processing logic - accepts a JSON request and returns a JSON response.
        """
        logger.info("Process request")
        g._elg_args = kwargs
        even_stream = True if "text/event-stream" in input_request.accept_mimetypes else False
        logger.debug("Accept MimeTypes: {mimetypes}", mimetypes=input_request.accept_mimetypes)
        logger.debug("Accept even-stream: {even_stream}", even_stream=even_stream)
        if input_request.mimetype == "multipart/form-data":
            data, generator = await self.parse_multipart()
            if self.consume_generator:
                data["content"] = await self.convert_async_generator_to_bytes(generator)
            else:
                data["generator"] = generator
        elif input_request.mimetype == "audio/mpeg":
            data, content = await self.parse_plain("MP3")
            data["content"] = content
        elif input_request.mimetype == "audio/wav" or input_request.mimetype == "audio/x-wav":
            data, content = await self.parse_plain("LINEAR16")
            data["content"] = content
        elif input_request.mimetype.startswith("image/"):
            mime_type = input_request.mimetype.split("/")[1]
            data, content = await self.parse_plain_image(mime_type)
            data["content"] = content
        elif input_request.mimetype == "application/json":
            data = await input_request.get_json()
        else:
            raise ProcessingError.UnsupportedType(input_request.content_type)

        logger.debug("Data type: {request_type}", request_type=data.get("type"))
        if data.get("type") == "audio":
            try:
                request = AudioRequest(**data)
            except Exception as e:
                raise ProcessingError.InvalidRequest()
        elif data.get("type") == "image":
            try:
                logger.info(data)
                request = ImageRequest(**data)
            except Exception as e:
                raise ProcessingError.InvalidRequest()
        elif data.get("type") == "text":
            try:
                request = TextRequest(**data)
            except Exception as e:
                raise ProcessingError.InvalidRequest()
        elif data.get("type") == "structuredText":
            try:
                request = StructuredTextRequest(**data)
            except Exception as e:
                raise ProcessingError.InvalidRequest()
        else:
            raise ProcessingError.InvalidRequest()
        logger.trace("Call with the input: {request}", request=request)
        logger.debug("Await for the coroutine...")
        try:
            response = await self.process_request(request)
        except ProcessingError:
            # if the code raises its own ProcessingError with a valid ELG error
            # message, just pass that along as-is rather than converting to a
            # generic internalError
            raise
        except:
            logger.opt(exception=True).warning("Exception during processing")
            raise ProcessingError.InternalError(str(sys.exc_info()[1]))
        if isinstance(response, Failure):
            logger.debug("Get error message")
            response = {"failure": response}
            logger.trace("Return: {response}", response=response)
            logger.info("Response returned")
            return response
        elif isinstance(response, ResponseObject):
            logger.debug("Get response")
            response = {"response": response}
            logger.trace("Return: {response}", response=response)
            logger.info("Response returned")
            return response
        elif isinstance(response, AsyncIterable):
            logger.debug("Get async iterable response")
            if even_stream:
                response = await make_response(
                    self.generator_mapping(response),
                    200,
                    {"Content-Type": "text/event-stream"},
                )
                # Quart will by default time-out long responses, may be necessary to disable that
                # or at least set a longer timeout than usual
                response.timeout = None
                logger.info("Streaming response returned")
                return response
            else:
                response = await self.get_response_from_generator(response)
                if isinstance(response, ResponseObject):
                    logger.debug("Get response")
                    response = {"response": response}
                elif isinstance(response, Failure):
                    logger.debug("Get error message")
                    response = {"failure": response}
                else:
                    logger.debug("Response type not known")
                    raise ProcessingError.InvalidRequest()
                logger.trace("Return: {response}", response=response)
                logger.info("Response returned")
                return response
        else:
            raise ProcessingError.InvalidRequest()

    async def process_request(self, request):
        """
        Method to process the request object. This method only calls the right process method regarding the type of the request.
        """
        if request.type == "text":
            logger.debug("Process text request")
            return await self.process_text(request)
        if request.type == "structuredText":
            logger.debug("Process structuredText request")
            return await self.process_structured_text(request)
        if request.type == "audio":
            logger.debug("Process audio request")
            return await self.process_audio(request)
        if request.type == "image":
            logger.debug("Process image request")
            return await self.process_image(request)
        raise ProcessingError.InvalidRequest()

    async def process_text(self, request: TextRequest):
        """
        Method to implement if the service takes text as input. This method must be implemented as async.

        Args:
            request (TextRequest): TextRequest object.
        """
        raise ProcessingError.UnsupportedType()

    async def process_structured_text(self, request: StructuredTextRequest):
        """
        Method to implement if the service takes structured text as input. This method must be implemented as async.

        Args:
            request (StructuredTextRequest): StructuredTextRequest object.
        """
        raise ProcessingError.UnsupportedType()

    async def process_audio(self, request: AudioRequest):
        """
        Method to implement if the service takes audio as input. This method must be implemented as async.

        Args:
            request (AudioRequest): AudioRequest object.
        """
        raise ProcessingError.UnsupportedType()

    async def process_image(self, request: ImageRequest):
        """
        Method to implement if the service takes an image as input. This method must be implemented as async.

        Args:
            request (ImageRequest): ImageRequest object.
        """
        raise ProcessingError.UnsupportedType()

    async def generator_mapping(self, generator):
        end = False
        try:
            async for message in generator:
                if end == True:
                    logger.warning(
                        "The service has already returned a ResponseObject or Failure message but continue to return a message. This message will be ignored and not returned to the user."
                    )
                    continue
                if isinstance(message, Failure):
                    logger.debug("Get failure message")
                    message = json.dumps({"failure": message}, cls=self.app.json_encoder)
                    end = True
                elif isinstance(message, Progress):
                    logger.debug("Get progress message")
                    message = json.dumps({"progress": message}, cls=self.app.json_encoder)
                elif isinstance(message, ResponseObject):
                    logger.debug("Get response")
                    message = json.dumps({"response": message}, cls=self.app.json_encoder)
                    end = True
                yield f"data:{message}\r\n\r\n".encode("utf-8")
        except:
            logger.opt(exception=True).warning("Exception in generator")
            message = json.dumps(
                {
                    "failure": Failure(
                        errors=[StandardMessages.generate_elg_service_internalerror(params=[str(sys.exc_info()[1])])]
                    )
                },
                cls=self.app.json_encoder,
            )
            yield f"data:{message}\r\n\r\n".encode("utf-8")

    @staticmethod
    async def get_response_from_generator(generator):
        response = None
        async for message in generator:
            if response is not None:
                logger.warning(
                    "The service has already returned a ResponseObject or Failure message but continue to return a message. This message will be ignored and not returned to the user."
                )
                continue
            if isinstance(message, (ResponseObject, Failure)):
                response = message
        if response is None:
            return Failure(errors=[StandardMessages.generate_elg_invalid_response()])
        return response

    @classmethod
    def create_requirements(cls, requirements: List = [], path: str = None):
        """
        Class method to create the correct requirements.txt file.

        Args:
            requirements (List, optional): List of required pip packages. Defaults to [].
            path (str, optional): Path where to generate the file. Defaults to None.
        """
        if path == None:
            path = Path(inspect.getsourcefile(cls))
        else:
            path = Path(path)
        requirements = cls.requirements + requirements
        with open(path.parent / "requirements.txt", "w") as f:
            f.write("\n".join(set(requirements)))

    @classmethod
    def create_docker_files(
        cls,
        required_files: List[str] = [],
        required_folders: List[str] = [],
        commands: List[str] = [],
        base_image: str = "python:3.8-slim",
        path: str = None,
        log_level: str = "INFO",
    ):
        """Class method to create the correct Dockerfile.

        Args:
            required_files (List[str], optional): List of files needed for the service. Defaults to [].
            required_folders (List[str], optional): List of folders needed for the service. Defaults to [].
            commands (List[str], optional): List off additional commands to run in the Dockerfile. Defaults to [].
            base_image (str, optional): Name of the base Docker image used in the Dockerfile. Defaults to 'python:3.8-slim'.
            path (str, optional): Path where to generate the file. Defaults to None.
            log_level (str, optional): The minimum severity level from which logged messages should be displayed. Defaults to 'INFO'.
        """
        if path == None:
            path = Path(inspect.getsourcefile(cls))
        else:
            path = Path(path)
        required_files = [path.name] + ["prepare.py"] + required_files
        required_folders = "\n".join(
            [COPY_FOLDER.format(folder_name=str(Path(folder))) for folder in required_folders]
        )
        commands = ["RUN " + cmd for cmd in commands]
        service_script = path.name[:-3]  # to remove .py
        # The docker-entrypoint file is a Linux shell script so _must_ be
        # written with Unix-style line endings, even if the build is being done
        # on Windows.  To ensure this we write in binary mode.
        with open(path.parent / "docker-entrypoint.sh", "wb") as f:
            f.write(ENTRYPOINT_QUART.format(service_script=service_script).encode("utf-8"))
        with open(path.parent / "prepare.py", "w") as f:
            f.write(
                f"""import asyncio
            
async def prepare():
    from {service_script} import app
    await app.startup()
    await app.shutdown()

asyncio.run(prepare())
"""
            )
        with open(path.parent / "Dockerfile", "w") as f:
            f.write(
                DOCKERFILE.format(
                    base_image=base_image,
                    required_files=" ".join(required_files),
                    required_folders=required_folders,
                    commands="\n".join(commands),
                    env=ENV_QUART.format(log_level=log_level),
                    caching=f"""RUN ["venv/bin/python", "prepare.py"]
ENV TRANSFORMERS_OFFLINE=1
""",
                )
            )

    @classmethod
    def docker_build_image(cls, tag: str, pull: bool = True, path: str = None, **kwargs):
        """
        Class method to do `docker build ...` in python. Better to use the docker cli instead of this method.
        """
        if path == None:
            path = Path(inspect.getsourcefile(cls))
        else:
            path = Path(path)
        client = docker.from_env()
        image, _ = client.images.build(path=str(path.parent), tag=tag, pull=pull, **kwargs)
        return image

    @classmethod
    def docker_push_image(
        cls,
        repository: str,
        tag: str,
        username: str = None,
        password: str = None,
        **kwargs,
    ):
        """
        Class method to do `docker push ...` in python. Better to use the docker cli instead of this method.
        """
        client = docker.from_env()
        if username is not None and password is not None:
            auth_config = {"username": username, "password": password}
            client.images.push(
                repository=repository,
                tag=tag,
                auth_config=auth_config,
                stream=True,
                **kwargs,
            )
        client.images.push(repository=repository, tag=tag, stream=True, **kwargs)
        return

    @classmethod
    def docker_build_push_image(
        cls,
        repository: str,
        tag: str,
        pull: bool = True,
        username: str = None,
        password: str = None,
        build_kwargs: Dict = {},
        push_kwargs: Dict = {},
    ):
        cls.docker_build_image(tag=f"{repository}:{tag}", pull=pull, **build_kwargs)
        cls.docker_push_image(
            repository=repository,
            tag=tag,
            username=username,
            password=password,
            **push_kwargs,
        )
        return None



from datetime import date, datetime, time
from json import JSONEncoder
from uuid import UUID

from pydantic import BaseModel

try:
    import collections.abc as collections_abc  # python 3.3+
except ImportError:
    import collections as collections_abc

try:
    from dataclasses import asdict, is_dataclass
except ImportError:
    # dataclasses not available
    is_dataclass = lambda x: False
    asdict = lambda x: dict()


def json_encoder(service):
    """Create a JSON encoder class that calls the `to_json` hook of the given FlaskService/QuartService"""

    class ELGJsonEncoder(JSONEncoder):
        def default(self, o):
            # First try the service hook
            val = service.to_json(o)
            if val is not None:
                return val
            # Hook did not return a substitute, so fall back on default behaviour
            elif isinstance(o, BaseModel):
                return o.dict(by_alias=True, exclude_none=True)
            elif isinstance(o, collections_abc.Iterable):
                # All iterables will be converted to list.
                return list(o)
            elif isinstance(o, (datetime, date, time)):
                return o.isoformat()
            elif isinstance(o, UUID):
                return str(o)
            elif is_dataclass(o):
                return asdict(o)
            elif hasattr(o, "__json__"):
                return o.__json__()
            elif hasattr(o, "for_json"):
                return o.for_json()
            else:
                return super().default(o)

    return ELGJsonEncoder



import json
import random
import re
import string
from pathlib import Path
from typing import Any, List, Union
from urllib.parse import urlparse

from loguru import logger

from .entity import Entity, MetadataRecordObj
from .utils import get_information
from .utils.local_installation import (DEFAULT_CONF_TEMPLATE, DOCKER_COMPOSE,
                                       EXPOSE_PORT_CONFIG, FRONTEND, GUI,
                                       GUI_CONF_TEMPLATE, HELPERS,
                                       HELPERS_TEMPLATE,
                                       HTML_INDEX_HTML_TEMPLATE,
                                       HTML_INDEX_IFRAME, HTML_INDEX_SCRIPT,
                                       I18N, I18N_CONF_TEMPLATE, LTSERVICE,
                                       LTSERVICE_SIDECAR, LTSERVICE_URL)


def _name_from_image(image):
    return re.sub("[^0-9a-zA-Z]+", "_", image)[-60:]


def _random_name(length: int = 10):
    return "".join(random.choice(string.ascii_uppercase + string.digits) for _ in range(length))


class LTServiceLocalInstallation:
    """
    Class that contains all the information to deploy an ELG-compatible service locally
    """

    def __init__(
        self,
        id: int,
        image: str,
        sidecar_image: str,
        name: str,
        image_envvars: List[str],
        full_name: str,
        port: int,
        path: str,
        gui: bool,
        gui_image: str,
        gui_port: int,
        gui_path: str,
        record: Any,
    ):
        """
        Initialize a LTServiceLocalInstallation object

        Args:
            id (int): id of the LT service
            image (str): docker image of the LT service
            sidecar_image (str): docker image of the sidecar container
            name (str): short name of the LT service used in the docker-compose file
            image_envvars (List[str]): environment variables to pass to the LT service docker container
            full_name (str): long name of the LT service used in the GUI
            port (int): port exposed by the LT service docker container
            path (str): path to the LT service endpoint
            gui (bool): boolean to indicate if yes or no the GUI should be deployed
            gui_image (str): docker image of the GUI
            gui_port (int): port exposed by the GUI docker container
            gui_path (str): path to the GUI endpoint
            record (Any): metadata record of the LT service
        """
        self.id = id
        self.image = image
        self.sidecar_image = sidecar_image
        self.name = name
        self.image_envvars = image_envvars
        self.full_name = full_name
        self.port = port
        self.path = path
        self.gui = gui
        self.gui_image = gui_image
        self.gui_port = gui_port
        self.gui_path = gui_path
        self.record = record

        self.ltservice = LTSERVICE.format(
            LTSERVICE_NAME=self.name,
            LTSERVICE_IMAGE=self.image,
            LTSERVICE_ENVVARS=str(self.image_envvars),
        )

        self.sidecarservice = (
            LTSERVICE_SIDECAR.format(LTSERVICE_NAME=self.name, SIDECAR_IMAGE=self.sidecar_image)
            if self.sidecar_image
            else ""
        )

        self.url = LTSERVICE_URL.format(
            LTSERVICE_NAME=self.name,
            LTSERVICE_PORT=self.port,
            LTSERVICE_PATH=self.path,
        )
        if self.gui:
            self.iframe = HTML_INDEX_IFRAME.format(
                LTSERVICE_ID=self.id,
                LTSERVICE_FULL_NAME=self.full_name,
                LTSERVICE_NAME=self.name,
            )
            self.script = HTML_INDEX_SCRIPT.format(
                LTSERVICE_NAME=self.name,
                GUI_NAME=_name_from_image(self.gui_image),
                GUI_PATH=self.gui_path,
            )

    @classmethod
    def from_id(
        cls,
        id: int,
        gui: bool = True,
        gui_image: str = "registry.gitlab.com/european-language-grid/usfd/gui-ie:latest",
        gui_port: int = 80,
        image_envvars: List[str] = [],
        domain: str = "live",
        use_cache: bool = True,
        cache_dir: str = "~/.cache/elg",
    ):
        """
        Class method to init a LTServiceLocalInstallation object from the id of an LT service deployed in the ELG cluster

        Args:
            id (int): id of the LT service in the ELG cluster
            gui (bool, optional): boolean to indicate if yes or no the GUI should be deployed. Defaults to True.
            gui_image (_type_, optional): docker image of the GUI. Defaults to "registry.gitlab.com/european-language-grid/usfd/gui-ie:latest".
            gui_port (int, optional): port exposed by the GUI docker container. Defaults to 80.
            image_envvars (List[str], optional): environment variables to pass to the LT service docker container. Defaults to [].
            domain (str, optional): domain of the ELG cluster. Defaults to "live".
            use_cache (bool, optional): True if you want to use cached files. Defaults to True.
            cache_dir (str, optional): path to the cache_dir. Set it to None to not store any cached files. Defaults to "~/.cache/elg".

        Returns:
            LTServiceLocalInstallation: the LTServiceLocalInstallation object created
        """
        entity = Entity.from_id(id=id, domain=domain, use_cache=use_cache, cache_dir=cache_dir)
        software_distribution = get_information(
            id=entity.id,
            obj=entity.record,
            infos=["described_entity", "lr_subclass", "software_distribution"],
        )[0]
        if software_distribution.get("private_resource"):
            logger.warning(
                "Service [{id}] is private. It cannot be deployed locally.",
                id=entity.id,
            )
            return None
        sidecar_image = software_distribution.get("service_adapter_download_location")
        image = software_distribution.get("docker_download_location")
        execution_location = urlparse(software_distribution.get("execution_location"))
        name = get_information(
            id=entity.id,
            obj=entity.record,
            infos=["service_info", "elg_execution_location_sync"],
        ).split("/")[-1]
        full_name = entity.resource_name
        port = execution_location.port
        path = execution_location.path
        gui_path = get_information(id=entity.id, obj=entity.record, infos=["service_info", "elg_gui_url"],).split(
            "/"
        )[-1]
        record = entity.record
        return cls(
            id=id,
            image=image,
            sidecar_image=sidecar_image,
            name=name,
            image_envvars=image_envvars,
            full_name=full_name,
            port=port,
            path=path,
            gui=gui,
            gui_image=gui_image,
            gui_port=gui_port,
            gui_path=gui_path,
            record=record,
        )

    @classmethod
    def from_docker_image(
        cls,
        image: str,
        execution_location: str,
        sidecar_image: str = "",
        name: str = None,
        image_envvars: List[str] = [],
        full_name: str = None,
        gui: bool = False,
        gui_image: str = "registry.gitlab.com/european-language-grid/usfd/gui-ie:latest",
        gui_port: int = 80,
        gui_path: str = "",
        record: Any = {},
    ):
        """
        Class method to init a LTServiceLocalInstallation object from a docker image

        Args:
            image (str): docker image of the LT service
            execution_location (str): execution location of the LT service in the LT service docker container (e.g. http://localhost:8000/process)
            sidecar_image (str, optional): docker image of the sidecar. Defaults to "".
            name (str, optional): short name of the LT service used in the docker-compose. Defaults to None.
            image_envvars (List[str], optional): environment variables to pass to the LT service docker container. Defaults to [].
            full_name (str, optional): long name of the LT service used in the GUI. Defaults to None.
            gui (bool, optional): boolean to indicate if yes or no the GUI should be deplo. Defaults to False.
            gui_image (str, optional): docker image of the GUI. Defaults to "registry.gitlab.com/european-language-grid/usfd/gui-ie:latest".
            gui_port (int, optional): port exposed by the GUI docker container. Defaults to 80.
            gui_path (str, optional): path to the GUI endpoint. Defaults to "".
            record (Any, optional): metadata record of the LT service. Defaults to {}.

        Returns:
            LTServiceLocalInstallation: the LTServiceLocalInstallation object created
        """
        name = name if name else _random_name()
        full_name = full_name if full_name else f"ELG Service from Docker {name}"
        execution_location = urlparse(execution_location)
        port = execution_location.port
        port = int(port) if port is not None else 80
        path = execution_location.path
        return cls(
            id=-1,
            image=image,
            sidecar_image=sidecar_image,
            name=name,
            image_envvars=image_envvars,
            full_name=full_name,
            port=port,
            path=path,
            gui=gui,
            gui_image=gui_image,
            gui_port=gui_port,
            gui_path=gui_path,
            record=MetadataRecordObj(record),
        )


class LocalInstallation:
    """
    Class that contains all the information to deploy ELG-compatible services with a small part of the ELG infrastructure locally
    """

    def __init__(self, ltservices: List[LTServiceLocalInstallation], helpers: List[str] = None):
        """
        Initialize a LocalInstallation object

        Args:
            ltservices (List[LTServiceLocalInstallation]): list of the LT services local installation to deploy
            helpers (List[str], optional): list of helpers to deploy. Currently only "temp-storage" is a valid helper and can be used to deploy a temporary storage needed for some services. Defaults to None.
        """
        self.ltservices = ltservices
        self.helpers = helpers

    @classmethod
    def from_ids(
        cls,
        ids: List[int],
        gui: bool = True,
        gui_images: Union[str, List[str]] = "registry.gitlab.com/european-language-grid/usfd/gui-ie:latest",
        gui_ports: Union[int, List[int]] = 80,
        helpers: List[str] = None,
        images_envvars: List[List[str]] = None,
        domain: str = "live",
        use_cache: bool = True,
        cache_dir: str = "~/.cache/elg",
    ):
        """
        Class method to init a LocalInstallation object from multiple ids of LT services deployed in the ELG cluster

        Args:
            ids (List[int]): list of ids od LT services in the ELG cluster
            gui (bool, optional): boolean to indicate if yes or no the GUI should be deployed. Defaults to True.
            gui_images (Union[str, List[str], optional): docker images of the GUI. If string, the same GUI docker image will be used for all the services. Otherwise, the number of GUI docker images needs to be the same as the number of LT services deployed. Defaults to "registry.gitlab.com/european-language-grid/usfd/gui-ie:latest".
            gui_ports (Union[int, List[int]], optional): port exposed by the GUI docker container. If integer, the same port will be used for all the services. Otherwise, the number of port needs to be the same as the number of LT services deployed. Defaults to 80.
            helpers (List[str], optional): list of helpers to deploy. Currently only "temp-storage" is a valid helper and can be used to deploy a temporary storage needed for some services. Defaults to None.
            images_envvars (List[List[str]], optional): environment variables to pass to the LT services docker container. Defaults to None.
            domain (str, optional): domain of the ELG cluster. Defaults to "live".
            use_cache (bool, optional): True if you want to use cached files. Defaults to True.
            cache_dir (str, optional): path to the cache_dir. Set it to None to not store any cached files. Defaults to "~/.cache/elg".

        Returns:
            LocalInstallation: the LocalInstallation object created
        """
        if isinstance(gui_images, list) and len(gui_images) == 1:
            gui_images = gui_images[0]
        if isinstance(gui_images, str):
            gui_images = [gui_images for _ in range(len(ids))]
        elif isinstance(gui_images, list):
            if len(gui_images) != len(ids):
                raise ValueError(
                    f"You provided {len(gui_images)} GUI images and {len(ids)} service ids. These two numbers must be equal."
                )
        else:
            raise ValueError("gui_images must be a string or a list of strings.")

        if isinstance(gui_ports, list) and len(gui_ports) == 1:
            gui_ports = gui_ports[0]
        if isinstance(gui_ports, int):
            gui_ports = [gui_ports for _ in range(len(ids))]
        elif isinstance(gui_ports, list):
            if len(gui_ports) != len(ids):
                raise ValueError(
                    f"You provided {len(gui_ports)} GUI ports and {len(ids)} service ids. These two numbers must be equal."
                )
        else:
            raise ValueError("gui_ports must be a int or a list of ints.")

        if images_envvars is None:
            images_envvars = [[] for _ in range(len(ids))]
        if isinstance(images_envvars, list):
            if len(images_envvars) != len(ids):
                raise ValueError(
                    f"You provided {len(images_envvars)} images env variables and {len(ids)} service ids. These two numbers must be equal."
                )
        else:
            raise ValueError("images_envvars must be a list of the same length of the number of service ids.")

        assert len(ids) == len(gui_images)
        assert len(ids) == len(gui_ports)
        assert len(ids) == len(images_envvars)

        ltservices = [
            LTServiceLocalInstallation.from_id(
                id=ltservice_id,
                gui=gui,
                gui_image=gui_image,
                gui_port=gui_port,
                image_envvars=image_envvars,
                domain=domain,
                use_cache=use_cache,
                cache_dir=cache_dir,
            )
            for ltservice_id, gui_image, gui_port, image_envvars in zip(ids, gui_images, gui_ports, images_envvars)
        ]
        ltservices = [ltservice for ltservice in ltservices if ltservice]
        return cls(ltservices=ltservices, helpers=helpers)

    def create_docker_compose(
        self,
        expose_port: int = 8080,
        path: str = "./elg_local_installation/",
    ):
        """
        Method to generate the docker compose file and all the configuration files to deploy the LocalInstallation

        Args:
            expose_port (int, optional): port used to expose the GUI or the LT service execution server. Defaults to 8080.
            path (str, optional): path where to store the configuration files. Defaults to "./elg_local_installation/".
        """
        if self.ltservices == []:
            logger.warning("None of the services can be deployed locally. Therefore, no files will be created.")
            return
        guis_image_port = list(
            set([(ltservice.gui_image, ltservice.gui_port) for ltservice in self.ltservices if ltservice.gui])
        )
        gui = len(guis_image_port) > 0
        has_helpers = self.helpers and len(self.helpers) > 0
        if gui or has_helpers:
            guis = (
                (
                    [
                        GUI.format(GUI_NAME=_name_from_image(gui_image), GUI_IMAGE=gui_image)
                        for gui_image, _ in guis_image_port
                    ]
                    + [I18N]
                )
                if gui
                else []
            )
            helpers = []
            helpers_conf = []
            if has_helpers:
                for h in self.helpers:
                    if h in HELPERS:
                        helpers.append(HELPERS[h].format(EXPOSE_PORT=expose_port))
                        helpers_conf.append(HELPERS_TEMPLATE[h])
                    else:
                        logger.warning(f"Unknown helper {h}, ignored")
            frontend = FRONTEND.format(EXPOSE_PORT=expose_port)
            gui_conf_templates = (
                (
                    [
                        GUI_CONF_TEMPLATE.format(GUI_NAME=_name_from_image(gui_image), GUI_PORT=gui_port)
                        for gui_image, gui_port in guis_image_port
                    ]
                    + [I18N_CONF_TEMPLATE]
                )
                if gui
                else []
            )
            default_conf_template = DEFAULT_CONF_TEMPLATE.format(
                GUIS="\n\n".join(gui_conf_templates), HELPERS="\n\n".join(helpers_conf)
            )
            html_index_html_template = HTML_INDEX_HTML_TEMPLATE.format(
                IFRAMES="\n".join([ltservice.iframe for ltservice in self.ltservices if ltservice.gui]),
                SCRIPTS="\n".join([ltservice.script for ltservice in self.ltservices if ltservice.gui]),
            )

        docker_compose = DOCKER_COMPOSE.format(
            LTSERVICES="\n\n".join([ltservice.ltservice for ltservice in self.ltservices]),
            LTSERVICES_SIDECAR="\n".join(
                [ltservice.sidecarservice for ltservice in self.ltservices if ltservice.sidecarservice]
            ),
            LTSERVICES_URL="\n".join([ltservice.url for ltservice in self.ltservices]),
            EXPOSE_PORT=expose_port,
            EXPOSE_PORT_CONFIG=EXPOSE_PORT_CONFIG.format(EXPOSE_PORT=expose_port)
            if (not gui and not has_helpers)
            else "",
            GUIS="\n".join(guis) if gui else "",
            HELPERS="\n".join(helpers) if has_helpers else "",
            FRONTEND=frontend if gui or has_helpers else "",
        )
        path = Path(path)
        path.mkdir(parents=True, exist_ok=True)
        with open(path / "docker-compose.yml", "w") as f:
            f.write(docker_compose)
        if gui:
            nginx_conf_path = Path(path / "nginx-conf")
            nginx_conf_path.mkdir(parents=True, exist_ok=True)
            html_path = Path(nginx_conf_path / "html")
            html_path.mkdir(parents=True, exist_ok=True)
            with open(nginx_conf_path / "default.conf.template", "w") as f:
                f.write(default_conf_template)
            with open(html_path / "index.html.template", "w") as f:
                f.write(html_index_html_template)
            records_path = Path(nginx_conf_path / "records")
            records_path.mkdir(parents=True, exist_ok=True)
            for ltservice in self.ltservices:
                if ltservice.gui:
                    with open(records_path / f"{ltservice.name}.json", "w") as f:
                        json.dump(ltservice.record, f, default=lambda o: o.__dict__, indent=4)

        logger.info(
            "The Docker compose file has been created [{path}]. Run `docker-compose up` from the folder of Docker compose file to start the ELG REST server and the service.s.",
            path=path / "docker-compose.yml",
        )
        return



def to_camel(string: str) -> str:
    """
    Convert `snake_case` string to `lowerCamelCase`.
    """
    return "".join(word.capitalize() if i != 0 else word for i, word in enumerate(string.split("_")))



from collections.abc import AsyncIterable, Iterable
from pathlib import Path
from typing import Any, Dict, List

from pydantic import validator

from .. import Annotation, Request


class AudioRequest(Request):
    """
    Request representing a piece of audio - the actual audio data will be sent as a separate request.
    Subclass of :class:`elg.model.base.Request.Request`
    """

    type: str = "audio"
    """*(required)* the type of request must be \"audio\""""

    content: bytes = None
    """*(optional)* audio itself, if not being sent as separate stream"""

    generator: Any = None
    """*(optional)* generator that provide the audio itself"""

    format: str = "LINEAR16"
    """*(required)* format of audio request. must be either \"LINEAR16\" (default) or \"MP3\""""

    sample_rate: int = None
    """*(optional)* sample rate of audio"""

    features: Dict = None
    """*(optional)* arbitrary json metadata about content"""

    annotations: Dict[str, List[Annotation]] = None
    """*(optional)* optional annotations on request"""

    @validator("format")
    def format_must_be_specific(cls, v):
        """
        *(validator)* ensures the format of the audio request is either "LINEAR16" or "MP3"
        """
        acceptable_formats = ["LINEAR16", "MP3"]
        v = v.upper()
        if v not in acceptable_formats:
            raise ValueError("The format given is not supported")
        return v

    @validator("generator")
    def generator_must_be_iterable(cls, v):
        """
        *(validator)* ensures the iterator field of the audio request is either None or an Iterable
        """
        if v is None:
            return v
        if isinstance(v, (AsyncIterable, Iterable)):
            return v
        raise ValueError(f"The generator musts be None or an Iterable, not {type(v)}")

    @staticmethod
    def generator_from_file(filename, blocksize=1024):
        with open(filename, "rb") as file:
            byte = file.read(blocksize)
            while byte:
                yield byte
                byte = file.read(blocksize)

    @classmethod
    def from_file(
        cls,
        filename,
        format: str = None,
        sample_rate: int = None,
        features: Dict = None,
        annotations: Dict[str, List[Annotation]] = None,
        params: dict = None,
        streaming: bool = False,
        blocksize: int = 1024,
    ):
        """
        allows you to generate audio request from file
        """
        filename = Path(filename)
        if not filename.is_file():
            raise ValueError(f"{filename} musts be the path to a file.")
        if streaming:
            generator = cls.generator_from_file(filename=filename, blocksize=blocksize)
            content = None
        else:
            with open(filename, "rb") as f:
                content = f.read()
            generator = None
        if format is None:
            format = "MP3" if filename.suffix == ".mp3" else "LINEAR16"
        return cls(
            content=content,
            generator=generator,
            format=format,
            sample_rate=sample_rate,
            features=features,
            annotations=annotations,
            params=params,
        )

    @classmethod
    def from_auto_content(cls, content: Any):
        if isinstance(content, dict):
            return cls(**content)
        else:
            raise ValueError(f"The AudioRequest message cannot be initialize from {content}")

    def __str__(self):
        return " - ".join(
            [f"{k}: {v}" for k, v in self.dict(exclude={"content", "generator"}).items() if v is not None]
        ) + ((" - content " + str(len(self.content))) if self.content else (" - content generator"))



from base64 import b64decode
from pathlib import Path
from typing import Dict, List

from pydantic import validator

from .. import Annotation, ResponseObject


class AudioResponse(ResponseObject):
    """
    Response representing audio data with optional standoff annotations (e.g. text-to-speech results)
    Subclass of :class:`elg.model.base.ResponseObject.ResponseObject`

    """

    type: str = "audio"
    """*(required)*  type of response"""

    content: str
    """*(required)*  base64 encoded audio for short audio snippets"""

    format: str
    """*(required)*  specifies audio format used: either \"LINEAR16\" or \"MP3\""""

    features: dict = None
    """*(required)*  arbitrary json metadata about content"""

    annotations: Dict[str, List[Annotation]] = None
    """*(required)*  optional annotations on response"""

    @validator("format")
    def format_must_be_specific(cls, v):
        """
        *(validator)* ensures the format of the audio response is either \"LINEAR16\" or \"MP3\"
        """
        acceptable_formats = ["LINEAR16", "MP3"]
        if v.upper() not in acceptable_formats:
            raise ValueError(f"The format given ({v}) is not supported")
        return v.upper()

    def to_file(self, filename):
        """
        *(validator)* writes audio response to file
        """
        filename = Path(filename)
        with open(filename, "wb") as f:
            f.write(b64decode(self.content))

    def auto_content(self):
        if self.warnings is not None:
            return self.dict()
        if self.features is not None:
            return self.dict()
        if self.annotations is not None:
            return self.dict()
        return {
            "content": self.dict()["content"],
            "format": self.dict()["format"],
        }



ENTRYPOINT_FLASK = """\
#!/bin/sh\nexec /sbin/tini -- venv/bin/gunicorn --bind=0.0.0.0:8000 "--workers=$WORKERS" "--timeout=$TIMEOUT" "--worker-class=$WORKER_CLASS" --worker-tmp-dir=/dev/shm "$@" {service_script}:app\n"""

ENTRYPOINT_QUART = """\
#!/bin/sh\nexec /sbin/tini -- venv/bin/hypercorn --bind=0.0.0.0:8000 "$@" {service_script}:app\n"""

DOCKERFILE = """\
FROM {base_image}

# Install tini and create an unprivileged user
ADD https://github.com/krallin/tini/releases/download/v0.19.0/tini /sbin/tini
RUN addgroup --gid 1001 "elg" && adduser --disabled-password --gecos "ELG User,,," --home /elg --ingroup elg --uid 1001 elg && chmod +x /sbin/tini

# Copy in just the requirements file
COPY --chown=elg:elg requirements.txt /elg/

# Everything from here down runs as the unprivileged user account
USER elg:elg

WORKDIR /elg

{commands}

# Create a Python virtual environment for the dependencies
RUN python -m venv venv 
RUN /elg/venv/bin/python -m pip install --upgrade pip
RUN venv/bin/pip --no-cache-dir install -r requirements.txt 

# Copy ini the entrypoint script and everything else our app needs
{required_folders}
COPY --chown=elg:elg docker-entrypoint.sh {required_files} /elg/

# Many Python libraries used for LT such as nltk, transformers, etc. default to
# downloading their models from the internet the first time they are accessed.
# This is a problem for container images, as every run is the "first time"
# starting from a clean copy of the image.  Therefore it is strongly
# recommended to pre-download any models that your code depends on during the
# build, so they are cached within the final image.  For example:
#
# RUN venv/bin/python -m nltk.downloader -d venv/share/nltk_data punkt
#
# RUN venv/bin/python -c "from transformers import DistilBertTokenizer" \\
#                     -c "DistilBertTokenizer.from_pretrained('bert-base-uncased')"

{env}

{caching}

RUN chmod +x ./docker-entrypoint.sh
ENTRYPOINT ["./docker-entrypoint.sh"]\
"""

COPY_FOLDER = "COPY --chown=elg:elg {folder_name} /elg/{folder_name}/"

ENV_FLASK = """\
ENV WORKERS={workers}
ENV TIMEOUT={timeout}
ENV WORKER_CLASS={worker_class}
ENV LOGURU_LEVEL={log_level}
"""

ENV_QUART = """\
ENV LOGURU_LEVEL={log_level}
"""



import sys
from argparse import ArgumentParser
from typing import List, Union

from loguru import logger

from . import BaseELGCommand


def docker_create_command_factory(args):
    return DockerCreateCommand(
        classname=args.classname,
        path=args.path,
        requirements=args.requirements,
        requirements_file=args.requirements_file,
        required_files=args.required_files,
        required_folders=args.required_folders,
        commands=args.commands,
        base_image=args.base_image,
        service_type=args.service_type,
        log_level=args.log_level,
        workers=args.workers,
        timeout=args.timeout,
        worker_class=args.worker_class,
    )


class DockerCreateCommand(BaseELGCommand):
    @staticmethod
    def register_subcommand(parser: ArgumentParser):
        info_parser = parser.add_parser("create", description="Create requirements.txt and Docker files.")
        info_parser.add_argument(
            "-n",
            "--classname",
            type=str,
            default=None,
            required=True,
            help="Name of the Service Class",
        )
        info_parser.add_argument(
            "-p",
            "--path",
            type=str,
            default="",
            required=None,
            help="Path the python script containing the Service Class definition from the current location",
        )
        info_parser.add_argument(
            "-i",
            "--base_image",
            type=str,
            default="python:3.8-slim",
            required=None,
            help="Name of the base Docker image used in the Dockerfile",
        )
        info_parser.add_argument(
            "-r",
            "--requirements",
            type=str,
            action="append",
            default=None,
            required=False,
            help="Pypi requirements for the Service Class, e.g. `nltk==3.5`",
        )
        info_parser.add_argument(
            "--requirements_file",
            type=str,
            default=None,
            required=False,
            help="Pypi requirements file, e.g. `requirements.txt`",
        )
        info_parser.add_argument(
            "-f",
            "--required_files",
            type=str,
            action="append",
            default=None,
            required=False,
            help="Name of the files from the directory of the Service Class",
        )
        info_parser.add_argument(
            "--required_folders",
            type=str,
            action="append",
            default=None,
            required=False,
            help="Name of the folders from the directory of the Service Class",
        )
        info_parser.add_argument(
            "--commands",
            type=str,
            action="append",
            default=None,
            required=False,
            help="Command to run in the Dockerfile",
        )
        info_parser.add_argument(
            "-t",
            "--service_type",
            type=str,
            default="auto",
            required=None,
            help="Type of service used. Can be 'auto' (the service type is discovered), 'flask', or 'quart'.",
            choices=["flask", "quart"],
        )
        info_parser.add_argument(
            "--log_level",
            type=str,
            default="INFO",
            required=None,
            help="The minimum severity level from which logged messages should be displayed.",
            choices=["INFO", "DEBUG", "TRACE"],
        )
        info_parser.add_argument(
            "--workers",
            type=int,
            default=1,
            required=None,
            help="Number of Gunicorn workers. Only used for the FlaskService class.",
        )
        info_parser.add_argument(
            "--timeout",
            type=int,
            default=30,
            required=None,
            help="Timeout value for the Gunicorn worker. Only used for the FlaskService class.",
        )
        info_parser.add_argument(
            "--worker_class",
            type=str,
            default="sync",
            required=None,
            help="Worker class value for the Gunicorn worker. Only used for the FlaskService class.",
            choices=["sync", "gthread"],
        )

        info_parser.set_defaults(func=docker_create_command_factory)

    def __init__(
        self,
        classname: str = None,
        path: str = "",
        requirements: Union[str, List[str]] = None,
        requirements_file: str = None,
        required_files: Union[str, List[str]] = None,
        required_folders: Union[str, List[str]] = None,
        commands: Union[str, List[str]] = None,
        base_image: str = None,
        service_type: str = "flask",
        log_level: str = "INFO",
        workers: int = 1,
        timeout: int = 30,
        worker_class: str = "sync",
    ):
        self._classname = classname
        self._path = path
        self._requirements = requirements
        self._requirements_file = requirements_file
        self._required_files = required_files
        self._required_folders = required_folders
        self._commands = commands
        self._base_image = base_image
        self._service_type = service_type
        self._log_level = log_level
        self._workers = workers
        self._timeout = timeout
        self._worker_class = worker_class

    def run(self):
        if self._service_type == "auto":
            with open(self._path) as f:
                script = f.read()
            if "FlaskService" in script:
                logger.info(
                    "flask has been found as the service_type. Please use `--service_type quart` to overwrite it."
                )
                self._service_type = "flask"
            elif "QuartService" in script:
                logger.info(
                    "quart has been found as the service_type. Please use `--service_type flask` to overwrite it."
                )
                self._service_type = "quart"
            else:
                raise ValueError("Neither FlaskService nor QuartService were found in the Python script.")

        if self._service_type == "flask":
            from ..flask_service import FlaskService
        elif self._service_type == "quart":
            from ..quart_service import QuartService

        try:
            logger.info(f"Creation of `requirements.txt`")
            if self._requirements is None:
                requirements = []
            elif isinstance(self._requirements, str):
                requirements = [self._requirements]
            else:
                requirements = self._requirements
            if self._requirements_file is None:
                pass
            else:
                with open(self._requirements_file) as f:
                    requirements += f.read().split("\n")
            if self._service_type == "flask":
                FlaskService.create_requirements(requirements=requirements, path=self._path)
            elif self._service_type == "quart":
                QuartService.create_requirements(requirements=requirements, path=self._path)
            logger.info(f"Creation of `Dockerfile` and `docker-entrypoint.sh`")
            if self._required_files is None:
                required_files = []
            elif isinstance(self._required_files, str):
                required_files = [self._required_files]
            else:
                required_files = self._required_files
            if self._required_folders is None:
                required_folders = []
            elif isinstance(self._required_folders, str):
                required_folders = [self._required_folders]
            else:
                required_folders = self._required_folders
            if self._commands is None:
                commands = []
            elif isinstance(self._commands, str):
                commands = [self._commands]
            else:
                commands = self._commands
            if self._service_type == "flask":
                FlaskService.create_docker_files(
                    required_files=required_files,
                    required_folders=required_folders,
                    commands=commands,
                    base_image=self._base_image,
                    path=self._path,
                    log_level=self._log_level,
                    workers=self._workers,
                    timeout=self._timeout,
                    worker_class=self._worker_class,
                )
            elif self._service_type == "quart":
                QuartService.create_docker_files(
                    required_files=required_files,
                    required_folders=required_folders,
                    commands=commands,
                    base_image=self._base_image,
                    path=self._path,
                    log_level=self._log_level,
                )
        except Exception as e:
            logger.error(f"Error during the creation of docker files - {e}")
            sys.exit(1)



from .base import (Annotation, Failure, Progress, Request, ResponseObject,
                   StandardMessages, StatusMessage)
from .request import (AudioRequest, ImageRequest, StructuredTextRequest,
                      TextRequest)
from .response import (AnnotationsResponse, AudioResponse, ClassesResponse,
                       ClassificationResponse, TextsResponse,
                       TextsResponseObject, get_response)



from typing import Any

from pydantic import BaseModel

from .utils import to_camel


class Request(BaseModel):
    """
    Representation of a service invocation request.
    Intended to be abstract, subclasses should be initiated with their specific type

    **Subclasses**

    * :class:`elg.model.request.AudioRequest`
    * :class:`elg.model.request.TextRequest`
    * :class:`elg.model.request.StructuredTextRequest`
    """

    type: str = None
    """*(required in subclass)* the type of request"""
    params: dict = None
    """*(optional)* vendor specific params, up to service implementor to decide how to interpret these """

    class Config:
        alias_generator = to_camel

    def __str__(self):
        """Override string to display dictionary"""
        return " - ".join([f"{k}: {v}" for k, v in self.dict().items() if v is not None])

    def json(self, **kwargs: Any) -> str:
        if "exclude_none" not in kwargs.keys():
            kwargs["exclude_none"] = True
        return super().json(**kwargs)

    def __getitem__(self, key):
        return getattr(self, key)

    def get(self, key, value: Any = None):
        if hasattr(self, key):
            return getattr(self, key)
        return value



from typing import Any

from pydantic import BaseModel

from .StatusMessage import StatusMessage
from .utils import to_camel


class Progress(BaseModel):
    """
    Details of an in progress task
    Some LT services can take a long time to process each request - likely useful to keep caller updated
    """

    percent: float
    """*(required)* completion percentage"""
    message: StatusMessage = None
    """*(optional)* message describing progress report"""

    class Config:
        alias_generator = to_camel

    def json(self, **kwargs: Any) -> str:
        if "exclude_none" not in kwargs.keys():
            kwargs["exclude_none"] = True
        return super().json(**kwargs)

    def __getitem__(self, key):
        return getattr(self, key)

    def get(self, key, value: Any = None):
        if hasattr(self, key):
            return getattr(self, key)
        return value



from .StatusMessage import StatusMessage


class StandardMessages:

    """
    This class provides easy access to the standard set of ELG status messages that are provided by default by
    the platform and should be fully translated in the ELG user interface. If you use codes other than these
    standard ones in your services then you should also try to contribute translations of your messages into as
    many languages as possible for the benefit of other ELG users.

    Implementation note: This class is auto-generated from elg-messages.properties - to add new message codes you
    should edit the property files, then run /utils/generate_standard_messages.py. Do not edit this class directly.
    """

    @classmethod
    def generate_elg_request_invalid(cls, params=[], detail={}, **kwargs):
        """Generate StatusMessage for code: elg.request.invalid"""
        if "lang" in kwargs:
            print(
                "Warning: lang is deprecated and will be ignored. The standard message 'text' will be in English, pass it to the i18n resolver to obtain it in another language."
            )
        code = "elg.request.invalid"
        text = "Invalid request message"
        return StatusMessage(code=code, text=text, params=params, detail=detail)

    @classmethod
    def generate_elg_request_missing(cls, params=[], detail={}, **kwargs):
        """Generate StatusMessage for code: elg.request.missing"""
        if "lang" in kwargs:
            print(
                "Warning: lang is deprecated and will be ignored. The standard message 'text' will be in English, pass it to the i18n resolver to obtain it in another language."
            )
        code = "elg.request.missing"
        text = "No request provided in message"
        return StatusMessage(code=code, text=text, params=params, detail=detail)

    @classmethod
    def generate_elg_request_type_unsupported(cls, params=[], detail={}, **kwargs):
        """Generate StatusMessage for code: elg.request.type.unsupported"""
        if "lang" in kwargs:
            print(
                "Warning: lang is deprecated and will be ignored. The standard message 'text' will be in English, pass it to the i18n resolver to obtain it in another language."
            )
        code = "elg.request.type.unsupported"
        text = "Request type {0} not supported by this service"
        return StatusMessage(code=code, text=text, params=params, detail=detail)

    @classmethod
    def generate_elg_request_property_unsupported(cls, params=[], detail={}, **kwargs):
        """Generate StatusMessage for code: elg.request.property.unsupported"""
        if "lang" in kwargs:
            print(
                "Warning: lang is deprecated and will be ignored. The standard message 'text' will be in English, pass it to the i18n resolver to obtain it in another language."
            )
        code = "elg.request.property.unsupported"
        text = "Unsupported property {0} in request"
        return StatusMessage(code=code, text=text, params=params, detail=detail)

    @classmethod
    def generate_elg_request_too_large(cls, params=[], detail={}, **kwargs):
        """Generate StatusMessage for code: elg.request.too.large"""
        if "lang" in kwargs:
            print(
                "Warning: lang is deprecated and will be ignored. The standard message 'text' will be in English, pass it to the i18n resolver to obtain it in another language."
            )
        code = "elg.request.too.large"
        text = "Request size too large"
        return StatusMessage(code=code, text=text, params=params, detail=detail)

    @classmethod
    def generate_elg_request_parameter_missing(cls, params=[], detail={}, **kwargs):
        """Generate StatusMessage for code: elg.request.parameter.missing"""
        if "lang" in kwargs:
            print(
                "Warning: lang is deprecated and will be ignored. The standard message 'text' will be in English, pass it to the i18n resolver to obtain it in another language."
            )
        code = "elg.request.parameter.missing"
        text = "Required parameter {0} missing from request"
        return StatusMessage(code=code, text=text, params=params, detail=detail)

    @classmethod
    def generate_elg_request_parameter_invalid(cls, params=[], detail={}, **kwargs):
        """Generate StatusMessage for code: elg.request.parameter.invalid"""
        if "lang" in kwargs:
            print(
                "Warning: lang is deprecated and will be ignored. The standard message 'text' will be in English, pass it to the i18n resolver to obtain it in another language."
            )
        code = "elg.request.parameter.invalid"
        text = 'Value "{1}" is not valid for parameter {0}'
        return StatusMessage(code=code, text=text, params=params, detail=detail)

    @classmethod
    def generate_elg_request_text_mimetype_unsupported(cls, params=[], detail={}, **kwargs):
        """Generate StatusMessage for code: elg.request.text.mimeType.unsupported"""
        if "lang" in kwargs:
            print(
                "Warning: lang is deprecated and will be ignored. The standard message 'text' will be in English, pass it to the i18n resolver to obtain it in another language."
            )
        code = "elg.request.text.mimeType.unsupported"
        text = "MIME type {0} not supported by this service"
        return StatusMessage(code=code, text=text, params=params, detail=detail)

    @classmethod
    def generate_elg_request_audio_format_unsupported(cls, params=[], detail={}, **kwargs):
        """Generate StatusMessage for code: elg.request.audio.format.unsupported"""
        if "lang" in kwargs:
            print(
                "Warning: lang is deprecated and will be ignored. The standard message 'text' will be in English, pass it to the i18n resolver to obtain it in another language."
            )
        code = "elg.request.audio.format.unsupported"
        text = "Audio format {0} not supported by this service"
        return StatusMessage(code=code, text=text, params=params, detail=detail)

    @classmethod
    def generate_elg_request_audio_samplerate_unsupported(cls, params=[], detail={}, **kwargs):
        """Generate StatusMessage for code: elg.request.audio.sampleRate.unsupported"""
        if "lang" in kwargs:
            print(
                "Warning: lang is deprecated and will be ignored. The standard message 'text' will be in English, pass it to the i18n resolver to obtain it in another language."
            )
        code = "elg.request.audio.sampleRate.unsupported"
        text = "Audio sample rate {0} not supported by this service"
        return StatusMessage(code=code, text=text, params=params, detail=detail)

    @classmethod
    def generate_elg_request_image_format_unsupported(cls, params=[], detail={}, **kwargs):
        """Generate StatusMessage for code: elg.request.image.format.unsupported"""
        if "lang" in kwargs:
            print(
                "Warning: lang is deprecated and will be ignored. The standard message 'text' will be in English, pass it to the i18n resolver to obtain it in another language."
            )
        code = "elg.request.image.format.unsupported"
        text = "Image format {0} not supported by this service"
        return StatusMessage(code=code, text=text, params=params, detail=detail)

    @classmethod
    def generate_elg_request_structuredtext_property_unsupported(cls, params=[], detail={}, **kwargs):
        """Generate StatusMessage for code: elg.request.structuredText.property.unsupported"""
        if "lang" in kwargs:
            print(
                "Warning: lang is deprecated and will be ignored. The standard message 'text' will be in English, pass it to the i18n resolver to obtain it in another language."
            )
        code = "elg.request.structuredText.property.unsupported"
        text = 'Unsupported property {0} in "texts" of structuredText request'
        return StatusMessage(code=code, text=text, params=params, detail=detail)

    @classmethod
    def generate_elg_response_invalid(cls, params=[], detail={}, **kwargs):
        """Generate StatusMessage for code: elg.response.invalid"""
        if "lang" in kwargs:
            print(
                "Warning: lang is deprecated and will be ignored. The standard message 'text' will be in English, pass it to the i18n resolver to obtain it in another language."
            )
        code = "elg.response.invalid"
        text = "Invalid response message"
        return StatusMessage(code=code, text=text, params=params, detail=detail)

    @classmethod
    def generate_elg_response_type_unsupported(cls, params=[], detail={}, **kwargs):
        """Generate StatusMessage for code: elg.response.type.unsupported"""
        if "lang" in kwargs:
            print(
                "Warning: lang is deprecated and will be ignored. The standard message 'text' will be in English, pass it to the i18n resolver to obtain it in another language."
            )
        code = "elg.response.type.unsupported"
        text = "Response type {0} not supported"
        return StatusMessage(code=code, text=text, params=params, detail=detail)

    @classmethod
    def generate_elg_response_property_unsupported(cls, params=[], detail={}, **kwargs):
        """Generate StatusMessage for code: elg.response.property.unsupported"""
        if "lang" in kwargs:
            print(
                "Warning: lang is deprecated and will be ignored. The standard message 'text' will be in English, pass it to the i18n resolver to obtain it in another language."
            )
        code = "elg.response.property.unsupported"
        text = "Unsupported property {0} in response"
        return StatusMessage(code=code, text=text, params=params, detail=detail)

    @classmethod
    def generate_elg_response_texts_property_unsupported(cls, params=[], detail={}, **kwargs):
        """Generate StatusMessage for code: elg.response.texts.property.unsupported"""
        if "lang" in kwargs:
            print(
                "Warning: lang is deprecated and will be ignored. The standard message 'text' will be in English, pass it to the i18n resolver to obtain it in another language."
            )
        code = "elg.response.texts.property.unsupported"
        text = 'Unsupported property {0} in "texts" of texts response'
        return StatusMessage(code=code, text=text, params=params, detail=detail)

    @classmethod
    def generate_elg_response_classification_property_unsupported(cls, params=[], detail={}, **kwargs):
        """Generate StatusMessage for code: elg.response.classification.property.unsupported"""
        if "lang" in kwargs:
            print(
                "Warning: lang is deprecated and will be ignored. The standard message 'text' will be in English, pass it to the i18n resolver to obtain it in another language."
            )
        code = "elg.response.classification.property.unsupported"
        text = 'Unsupported property {0} in "classes" of classification response'
        return StatusMessage(code=code, text=text, params=params, detail=detail)

    @classmethod
    def generate_elg_service_not_found(cls, params=[], detail={}, **kwargs):
        """Generate StatusMessage for code: elg.service.not.found"""
        if "lang" in kwargs:
            print(
                "Warning: lang is deprecated and will be ignored. The standard message 'text' will be in English, pass it to the i18n resolver to obtain it in another language."
            )
        code = "elg.service.not.found"
        text = "Service {0} not found"
        return StatusMessage(code=code, text=text, params=params, detail=detail)

    @classmethod
    def generate_elg_async_call_not_found(cls, params=[], detail={}, **kwargs):
        """Generate StatusMessage for code: elg.async.call.not.found"""
        if "lang" in kwargs:
            print(
                "Warning: lang is deprecated and will be ignored. The standard message 'text' will be in English, pass it to the i18n resolver to obtain it in another language."
            )
        code = "elg.async.call.not.found"
        text = "Async call {0} not found"
        return StatusMessage(code=code, text=text, params=params, detail=detail)

    @classmethod
    def generate_elg_permissions_quotaexceeded(cls, params=[], detail={}, **kwargs):
        """Generate StatusMessage for code: elg.permissions.quotaExceeded"""
        if "lang" in kwargs:
            print(
                "Warning: lang is deprecated and will be ignored. The standard message 'text' will be in English, pass it to the i18n resolver to obtain it in another language."
            )
        code = "elg.permissions.quotaExceeded"
        text = "Authorized quota exceeded"
        return StatusMessage(code=code, text=text, params=params, detail=detail)

    @classmethod
    def generate_elg_permissions_accessdenied(cls, params=[], detail={}, **kwargs):
        """Generate StatusMessage for code: elg.permissions.accessDenied"""
        if "lang" in kwargs:
            print(
                "Warning: lang is deprecated and will be ignored. The standard message 'text' will be in English, pass it to the i18n resolver to obtain it in another language."
            )
        code = "elg.permissions.accessDenied"
        text = "Access denied"
        return StatusMessage(code=code, text=text, params=params, detail=detail)

    @classmethod
    def generate_elg_permissions_accessmanagererror(cls, params=[], detail={}, **kwargs):
        """Generate StatusMessage for code: elg.permissions.accessManagerError"""
        if "lang" in kwargs:
            print(
                "Warning: lang is deprecated and will be ignored. The standard message 'text' will be in English, pass it to the i18n resolver to obtain it in another language."
            )
        code = "elg.permissions.accessManagerError"
        text = "Error in access manager: {0}"
        return StatusMessage(code=code, text=text, params=params, detail=detail)

    @classmethod
    def generate_elg_file_not_found(cls, params=[], detail={}, **kwargs):
        """Generate StatusMessage for code: elg.file.not.found"""
        if "lang" in kwargs:
            print(
                "Warning: lang is deprecated and will be ignored. The standard message 'text' will be in English, pass it to the i18n resolver to obtain it in another language."
            )
        code = "elg.file.not.found"
        text = "File {0} not found"
        return StatusMessage(code=code, text=text, params=params, detail=detail)

    @classmethod
    def generate_elg_file_expired(cls, params=[], detail={}, **kwargs):
        """Generate StatusMessage for code: elg.file.expired"""
        if "lang" in kwargs:
            print(
                "Warning: lang is deprecated and will be ignored. The standard message 'text' will be in English, pass it to the i18n resolver to obtain it in another language."
            )
        code = "elg.file.expired"
        text = "Requested file {0} no longer available"
        return StatusMessage(code=code, text=text, params=params, detail=detail)

    @classmethod
    def generate_elg_upload_too_large(cls, params=[], detail={}, **kwargs):
        """Generate StatusMessage for code: elg.upload.too.large"""
        if "lang" in kwargs:
            print(
                "Warning: lang is deprecated and will be ignored. The standard message 'text' will be in English, pass it to the i18n resolver to obtain it in another language."
            )
        code = "elg.upload.too.large"
        text = "Upload too large"
        return StatusMessage(code=code, text=text, params=params, detail=detail)

    @classmethod
    def generate_elg_service_internalerror(cls, params=[], detail={}, **kwargs):
        """Generate StatusMessage for code: elg.service.internalError"""
        if "lang" in kwargs:
            print(
                "Warning: lang is deprecated and will be ignored. The standard message 'text' will be in English, pass it to the i18n resolver to obtain it in another language."
            )
        code = "elg.service.internalError"
        text = "Internal error during processing: {0}"
        return StatusMessage(code=code, text=text, params=params, detail=detail)



from typing import Any, List

from pydantic import BaseModel

from .StatusMessage import StatusMessage
from .utils import to_camel


class ResponseObject(BaseModel):
    """
    Representation of a successful completion response.
    Abstract, subclasses must instantiate this with their own type

    **Subclasses**

    * :class:`elg.model.response.AnnotationsResponse`
    * :class:`elg.model.response.AudioResponse`
    * :class:`elg.model.response.ClassificationResponse`
    * :class:`elg.model.response.TextsResponse`
    """

    type: str
    """*(required in subclass)* the type of response"""

    warnings: List[StatusMessage] = None
    """*(optional)* messages describing any warnings on response"""

    class Config:
        alias_generator = to_camel

    def json(self, **kwargs: Any) -> str:
        if "exclude_none" not in kwargs.keys():
            kwargs["exclude_none"] = True
        return super().json(**kwargs)

    def __getitem__(self, key):
        return getattr(self, key)

    def get(self, key, value: Any = None):
        if hasattr(self, key):
            return getattr(self, key)
        return value



from numbers import Number
from typing import Any, Dict, List

from pydantic import BaseModel, root_validator

from .. import Annotation, ResponseObject
from ..base.utils import to_camel


class TextsResponseObject(BaseModel):
    """
    Object representing a structured piece of text. Recursive.
    """

    role: str = None
    """*(optional)* the role of this node in the response"""

    content: str = None
    """*(optional)* string of translated/transcribed text"""

    texts: "List[TextsResponseObject]" = None
    """*(optional)* list of same structures, recursive"""

    score: Number = None
    """*(optional)* confidence of response"""

    features: dict = None
    """*(optional)* arbitrary JSON metadata about content"""

    annotations: Dict[str, List[Annotation]] = None
    """*(optional)* optional annotations on request"""

    class Config:
        alias_generator = to_camel
        arbitrary_types_allowed = True

    @root_validator()
    def either_content_or_text(cls, values):
        """
        *(validator)* ensures either the \"content\" or \"text\" fields are present
        """
        content, texts, score = (
            values.get("content"),
            values.get("texts"),
            values.get("score"),
        )
        if content is None and texts is None:
            raise ValueError('A text response must have either "content" or "texts" fields')
        if texts and score:
            raise ValueError("A branch node should not have a score attached")
        return values

    def json(self, **kwargs: Any) -> str:
        if "exclude_none" not in kwargs.keys():
            kwargs["exclude_none"] = True
        return super().json(**kwargs)

    def __getitem__(self, key):
        return getattr(self, key)

    def get(self, key, value: Any = None):
        if hasattr(self, key):
            return getattr(self, key)
        return value

    def auto_content(self):
        if self.features is not None and self.annotations is not None:
            return self.dict()
        if self.content is not None and self.texts is None:
            if self.score is None:
                return self.content
            return {
                "content": self.content,
                "score": self.score,
            }
        if self.content is None and self.texts is not None:
            if len(self.texts) == 1:
                return self.texts[0].auto_content()
            return self.texts


TextsResponseObject.update_forward_refs()


class TextsResponse(ResponseObject):
    """
    Response consisting of a set of one or more new texts, each with optional annotations attached to it.
    Subclass of :class:`elg.model.base.ResponseObject.ResponseObject`

    For example a set of possible translations produced by a translation tool or possible transcriptions produced by a
    speech-to-text recogniser.
    """

    type: str = "texts"
    """*(optional)* type of response, must be \"texts\""""

    texts: List[TextsResponseObject]
    """*(optional)* list of objects representing a structured text response"""

    def auto_content(self):
        if self.warnings is not None:
            return self.dict()
        if len(self.texts) == 1:
            return self.texts[0].auto_content()
        return self.dict()["texts"]



from typing import Any, List

from pydantic import BaseModel, Field

from .. import ResponseObject
from ..base.utils import to_camel


class ClassesResponse(BaseModel):
    """
    Classification object: classification and score (optional likelihood of classification)
    Subclass of :class:`elg.model.base.ResponseObject.ResponseObject`
    """

    class_field: str = Field(None, alias="class")
    """*(required)* labelled class"""

    score: float = None
    """*(optional)* confidence score in class"""

    class Config:
        alias_generator = to_camel

    def json(self, **kwargs: Any) -> str:
        if "exclude_none" not in kwargs.keys():
            kwargs["exclude_none"] = True
        return super().json(**kwargs)

    def __getitem__(self, key):
        return getattr(self, key)

    def get(self, key, value: Any = None):
        if hasattr(self, key):
            return getattr(self, key)
        return value


class ClassificationResponse(ResponseObject):
    """
    Response encapsulating one or more classifications of the whole input message, optionally with confidence scores
    attached.
    """

    type: str = "classification"
    """*(required)* type of response"""

    classes: List[ClassesResponse] = None
    """*(optional)*  list of classifications, zero or more allowed"""

    def auto_content(self):
        if self.warnings is not None:
            return self.dict()
        if len(self.classes) == 1:
            return self.classes[0].dict()
        return self.dict()["classes"]



import sys
import traceback
from argparse import ArgumentParser
from typing import List

from loguru import logger

from . import BaseELGCommand


def local_installation_command_factory(args):
    return LocalInstallationIdsCommand(
        ids=args.ids,
        folder=args.folder,
        no_gui=args.no_gui,
        domain=args.domain,
        not_use_cache=args.not_use_cache,
        cache_dir=args.cache_dir,
        expose_port=args.expose_port,
        gui_images=args.gui_images,
        gui_ports=args.gui_ports,
        helpers=args.helper,
    )


class LocalInstallationIdsCommand(BaseELGCommand):
    @staticmethod
    def register_subcommand(parser: ArgumentParser):
        local_installation_parser = parser.add_parser(
            "ids",
            description="Create a Docker compose file to deploy a set of ELG services locally from their ids",
        )
        local_installation_parser.add_argument(
            "ids",
            type=int,
            nargs="+",
            help="ID of the service to deploy locally.",
        )
        local_installation_parser.add_argument(
            "--folder",
            type=str,
            default="./elg_local_installation/",
            required=False,
            help="path to the folder where to save the Docker compose file",
        )
        local_installation_parser.add_argument(
            "--no_gui",
            default=False,
            action="store_true",
            required=False,
            help="use to not use the GUI",
        )
        local_installation_parser.add_argument(
            "--domain",
            type=str,
            default="live",
            required=False,
            help="ELG domain you want to use",
        )
        local_installation_parser.add_argument(
            "--not_use_cache",
            default=False,
            action="store_true",
            required=False,
            help="use to not use cache",
        )
        local_installation_parser.add_argument(
            "--cache_dir",
            type=str,
            default="~/.cache/elg",
            required=False,
            help="path to the cache directory",
        )
        local_installation_parser.add_argument(
            "--expose_port",
            type=int,
            default=8080,
            required=False,
            help="port used to publish to the host",
        )
        local_installation_parser.add_argument(
            "--gui_images",
            type=str,
            nargs="+",
            default="registry.gitlab.com/european-language-grid/usfd/gui-ie:latest",
            required=False,
            help="docker image of the GUI",
        )
        local_installation_parser.add_argument(
            "--gui_ports",
            type=int,
            nargs="+",
            default=80,
            required=False,
            help="port used by the GUI docker image",
        )
        local_installation_parser.add_argument(
            "--helper",
            type=str,
            required=False,
            default=[],
            action="append",
            choices=["temp-storage"],
            help="helper services to include in the compose stack",
        )

        local_installation_parser.set_defaults(func=local_installation_command_factory)

    def __init__(
        self,
        ids: List[int] = None,
        folder: str = "./",
        no_gui: bool = False,
        domain: str = "live",
        not_use_cache: bool = False,
        cache_dir: str = "~/.cache/elg",
        expose_port: int = 8080,
        gui_images: str = "registry.gitlab.com/european-language-grid/usfd/gui-ie:latest",
        gui_ports: int = 80,
        helpers: List[str] = None,
    ):
        self._ids = ids
        self._folder = folder
        self._gui = not no_gui
        self._domain = domain
        self._use_cache = not not_use_cache
        self._cache_dir = cache_dir
        self._expose_port = expose_port
        self._gui_images = gui_images
        self._gui_ports = gui_ports
        self._helpers = helpers

    def run(self):
        from ..local_installation import LocalInstallation

        try:
            LocalInstallation.from_ids(
                ids=self._ids,
                gui=self._gui,
                gui_images=self._gui_images,
                gui_ports=self._gui_ports,
                helpers=self._helpers,
                domain=self._domain,
                use_cache=self._use_cache,
                cache_dir=self._cache_dir,
            ).create_docker_compose(
                expose_port=self._expose_port,
                path=self._folder,
            )
        except Exception as e:
            logger.error(f"Error during the creation of the Docker compose file - {e}")
            traceback.print_exc()
            sys.exit(1)



import sys
from argparse import ArgumentParser

from loguru import logger

from . import BaseELGCommand
from .local_installation_docker import LocalInstallationDockerCommand
from .local_installation_ids import LocalInstallationIdsCommand


class LocalInstallationCommand(BaseELGCommand):
    @staticmethod
    def register_subcommand(parser: ArgumentParser):
        commands_parser = parser.add_parser("local-installation")
        subparsers = commands_parser.add_subparsers(
            title="ELG Local Installation CLI",
            description="Create a Docker compose file to deploy a set of ELG services locally.",
            help="For more information about a command, run: `elg local-installation <command> --help`",
        )
        LocalInstallationIdsCommand.register_subcommand(subparsers)
        LocalInstallationDockerCommand.register_subcommand(subparsers)



from setuptools import find_packages, setup

setup(
    name="elg",
    version="0.4.22",
    author="ELG Technical Team",
    url="https://gitlab.com/european-language-grid/platform/python-client",
    author_email="contact@european-language-grid.eu",
    description="Use the European Language Grid in your Python projects",
    long_description=open("README.md", "r", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    keywords="tools, sdk, language technology, europe, european, nlp",
    license="MIT",
    packages=find_packages(exclude=("tests",)),
    install_requires=[
        "requests>=2.25",
        "tqdm>=4.49",
        "pandas>=1.2",
        "loguru>=0.5",
        "pydantic>=1.7.4",
    ],
    extras_require={
        "flask": [
            "flask>=2.0",
            "docker>=5.0",
            "requests_toolbelt>=0.9",
        ],
        "quart": [
            "quart>=0.15.1",
            "docker>=5.0",
            "requests_toolbelt>=0.9",
        ],
        "quality": ["black==21.4b0", "isort==5.5.4"],
    },
    entry_points={"console_scripts": ["elg=elg.cli.elg:main"]},
    python_requires=">=3.6.0",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
)



import inspect
import json
import sys
from collections.abc import Iterable
from pathlib import Path
from typing import Dict, List

from loguru import logger

try:
    import docker
    from flask import Flask, g
    from flask import request as flask_request
    from requests_toolbelt import MultipartDecoder
except:
    raise ImportError(
        "Extra dependencies need to be install to use the FlaskService class. Please run: `pip install elg[flask]`."
    )


from .model import (AudioRequest, Failure, ImageRequest, Progress, Request,
                    ResponseObject, StandardMessages, StructuredTextRequest,
                    TextRequest)
from .utils.docker import COPY_FOLDER, DOCKERFILE, ENTRYPOINT_FLASK, ENV_FLASK
from .utils.json_encoder import json_encoder


class FlaskService:
    """
    Class to help the creation of an ELG compatible service from a python model.
    Extra dependencies need to be install to use the FlaskService class. Please run: `pip install elg[flask]`.
    """

    requirements = ["gunicorn", "elg[flask]"]

    def __init__(self, name: str = "My ELG Service", path: str = "/process"):
        """
        Init function of the FlaskService

        Args:
            name (str, optional): Name of the service. It doesn't have any importance.
            path (str, optional): Path of the endpoint to expose. It can contain parameters, e.g., "/translate/<src>/<target>"
                later accessible in the process_* methods using the `self.url_param` method, e.g.,
                self.url_param('src'). Default to "/process".
        """
        self.name = name
        self.app = Flask(name)
        # Don't add an extra "status" property to JSON responses - this would break the API contract
        self.app.config["JSON_ADD_STATUS"] = False
        # Don't sort properties alphabetically in response JSON
        self.app.config["JSON_SORT_KEYS"] = False
        self.app.json_encoder = json_encoder(self)
        self.app.add_url_rule(path, "process", self.process, methods=["POST"])

    def to_json(self, obj):
        """Hook that can be overridden by subclasses to customise JSON encoding.

        FlaskService can convert the following types to JSON by default, in
        addition to the types handled natively by `json.dump`:

        - date, time, datetime (via `.isoformat()`)
        - uuid.UUID (converted to `str`)
        - any pydantic.BaseModel including the ELG message types (via
          `.dict(by_alias=True, exclude_none=True)`)
        - anything Iterable (as a list)
        - any `dataclass` (converted to a `dict` via `dataclasses.asdict`)
        - anything with a `__json__` or `for_json` method (which is expected to
          return a serializable type)

        To handle other types, or to change the standard behaviour for any of
        the above types, subclasses can override this method, which will be
        called whenever an object other than a string, number, bool, list or
        dict must be serialized and is expected to return a JSON-serializable
        object to be used in place of the original, or `None` to fall back to
        the default behaviour.

        The default implementation of this method always returns `None`.

        Args:
            obj: the object to convert

        Returns:
            a substitute object suitable for JSON serialization, or `None` to
            use the default behaviour.
        """
        return None

    def run(self):
        """
        Method to start the flask app.
        """
        self.app.run()

    def url_param(self, name: str):
        """
        Method to get give access to url parameters
        """
        return g._elg_args[name]

    def process(self, **kwargs):
        """
        Main request processing logic - accepts a JSON request and returns a JSON response.
        """
        logger.info("Process request")
        g._elg_args = kwargs
        even_stream = True if "text/event-stream" in flask_request.accept_mimetypes else False
        logger.debug("Accept MimeTypes: {mimetypes}", mimetypes=flask_request.accept_mimetypes)
        logger.debug("Accept even-stream: {even_stream}", even_stream=even_stream)
        if "application/json" in flask_request.content_type:
            data = flask_request.get_json()
        elif "multipart/form-data" in flask_request.content_type:
            decoder = MultipartDecoder(flask_request.get_data(), flask_request.content_type)
            data = {}
            for part in decoder.parts:
                headers = {k.decode(): v.decode() for k, v in part.headers.items()}
                if "application/json" in headers["Content-Type"]:
                    for k, v in json.loads(part.content.decode()).items():
                        data[k] = v
                elif "audio" in headers["Content-Type"] or "image" in headers["Content-Type"]:
                    data["content"] = part.content
                else:
                    raise ValueError("Unknown Content-Type in multipart request")
        else:
            raise ValueError()

        request_type = data.get("type")
        logger.debug("Data type: {request_type}", request_type=request_type)
        if request_type in ["audio", "image", "text", "structuredText"]:
            try:
                if request_type == "audio":
                    request = AudioRequest(**data)
                elif request_type == "image":
                    request = ImageRequest(**data)
                elif request_type == "text":
                    request = TextRequest(**data)
                else:
                    request = StructuredTextRequest(**data)
            except Exception as e:
                response = Failure(errors=[StandardMessages.generate_elg_request_invalid()])
        else:
            response = Failure(errors=[StandardMessages.generate_elg_request_invalid()])

        if isinstance(request, Request):
            logger.trace("Call with the input: {request}", request=request)
            try:
                response = self.process_request(request)
            except:
                logger.opt(exception=True).warning("Error during the request processing.")
                response = Failure(
                    errors=[StandardMessages.generate_elg_service_internalerror(params=[str(sys.exc_info()[1])])]
                )

        if isinstance(response, Failure):
            logger.debug("Get error message")
            response = {"failure": response}
            logger.trace("Return: {response}", response=response)
            logger.info("Response returned")
            return response
        elif isinstance(response, ResponseObject):
            logger.debug("Get response")
            response = {"response": response}
            logger.trace("Return: {response}", response=response)
            logger.info("Response returned")
            return response
        elif isinstance(response, Iterable):
            logger.debug("Get iterable response")
            if even_stream:
                logger.info("Streaming response returned")
                return self.app.response_class(self.generator_mapping(response), mimetype="text/event-stream")
            else:
                response = self.get_response_from_generator(response)
                if isinstance(response, ResponseObject):
                    logger.debug("Get response")
                    response = {"response": response}
                elif isinstance(response, Failure):
                    logger.debug("Get error message")
                    response = {"failure": response}
                else:
                    logger.debug("Response type not known")
                    response = {"failure": Failure(errors=[StandardMessages.generate_elg_invalid_response()])}
                logger.trace("Return: {response}", response=response)
                logger.info("Response returned")
                return response
        else:
            raise ValueError("Unknown returned type")

    def process_request(self, request):
        """
        Method to process the request object. This method only calls the right process method regarding the type of the request.
        """
        if request.type == "text":
            logger.debug("Process text request")
            return self.process_text(request)
        elif request.type == "structuredText":
            logger.debug("Process structuredText request")
            return self.process_structured_text(request)
        elif request.type == "image":
            logger.debug("Process image request")
            return self.process_image(request)
        elif request.type == "audio":
            logger.debug("Process audio request")
            return self.process_audio(request)
        return Failure(errors=[StandardMessages.generate_elg_request_invalid()])

    def process_text(self, request: TextRequest):
        """
        Method to implement if the service takes text as input.

        Args:
            request (TextRequest): TextRequest object.
        """
        raise NotImplementedError()

    def process_structured_text(self, request: StructuredTextRequest):
        """
        Method to implement if the service takes structured text as input.

        Args:
            request (StructuredTextRequest): StructuredTextRequest object.
        """
        raise NotImplementedError()

    def process_audio(self, request: AudioRequest):
        """
        Method to implement if the service takes audio as input.

        Args:
            request (AudioRequest): AudioRequest object.
        """
        raise NotImplementedError()

    def process_image(self, request: ImageRequest):
        """
        Method to implement if the service takes image as input.

        Args:
            request (ImageRequest): ImageRequest object.
        """
        raise NotImplementedError()

    def generator_mapping(self, generator):
        end = False
        try:
            for message in generator:
                if end == True:
                    logger.warning(
                        "The service has already returned a ResponseObject or Failure message but continue to return a message. This message will be ignored and not returned to the user."
                    )
                    continue
                if isinstance(message, Failure):
                    logger.debug("Get failure message")
                    message = json.dumps({"failure": message}, cls=self.app.json_encoder)
                    end = True
                elif isinstance(message, Progress):
                    logger.debug("Get progress message")
                    message = json.dumps({"progress": message}, cls=self.app.json_encoder)
                elif isinstance(message, ResponseObject):
                    logger.debug("Get response")
                    message = json.dumps({"response": message}, cls=self.app.json_encoder)
                    end = True
                yield f"data:{message}\r\n\r\n"
        except:
            logger.opt(exception=True).warning("Exception in generator")
            message = json.dumps(
                {
                    "failure": Failure(
                        errors=[StandardMessages.generate_elg_service_internalerror(params=[str(sys.exc_info()[1])])]
                    )
                },
                cls=self.app.json_encoder,
            )
            yield f"data:{message}\r\n\r\n"

    @staticmethod
    def get_response_from_generator(generator):
        response = None
        for message in generator:
            if response is not None:
                logger.warning(
                    "The service has already returned a ResponseObject or Failure message but continue to return a message. This message will be ignored and not returned to the user."
                )
                continue
            if isinstance(message, (ResponseObject, Failure)):
                response = message
        if response is None:
            return Failure(errors=[StandardMessages.generate_elg_invalid_response()])
        return response

    @classmethod
    def create_requirements(cls, requirements: List = [], path: str = None):
        """
        Class method to create the correct requirements.txt file.

        Args:
            requirements (List, optional): List of required pip packages. Defaults to [].
            path (str, optional): Path where to generate the file. Defaults to None.
        """
        if path == None:
            path = Path(inspect.getsourcefile(cls))
        else:
            path = Path(path)
        requirements = cls.requirements + requirements
        with open(path.parent / "requirements.txt", "w") as f:
            f.write("\n".join(set(requirements)))

    @classmethod
    def create_docker_files(
        cls,
        required_files: List[str] = [],
        required_folders: List[str] = [],
        commands: List[str] = [],
        base_image: str = "python:3.8-slim",
        path: str = None,
        log_level: str = "INFO",
        workers: int = 1,
        timeout: int = 30,
        worker_class: str = "sync",
    ):
        """Class method to create the correct Dockerfile.

        Args:
            required_files (List[str], optional): List of files needed for the service. Defaults to [].
            required_folders (List[str], optional): List of folders needed for the service. Defaults to [].
            commands (List[str], optional): List off additional commands to run in the Dockerfile. Defaults to [].
            base_image (str, optional): Name of the base Docker image used in the Dockerfile. Defaults to 'python:3.8-slim'.
            path (str, optional): Path where to generate the file. Defaults to None.
            log_level (str, optional): The minimum severity level from which logged messages should be displayed. Defaults to 'INFO'.
            workers (int, optional): Number of Gunicorn workers. Defaults to 1.
            timeout (int, optional): Timeout value for the Gunicorn worker. Defaults to 30.
            worker_class (str, optional): Worker class value for the Gunicorn worker. Defaults to 'sync'.
        """
        if path == None:
            path = Path(inspect.getsourcefile(cls))
        else:
            path = Path(path)
        required_files = [path.name] + required_files
        required_folders = "\n".join(
            [COPY_FOLDER.format(folder_name=str(Path(folder))) for folder in required_folders]
        )
        commands = ["RUN " + cmd for cmd in commands]
        service_script = path.name[:-3]  # to remove .py
        # The docker-entrypoint file is a Linux shell script so _must_ be
        # written with Unix-style line endings, even if the build is being done
        # on Windows.  To ensure this we write in binary mode.
        with open(path.parent / "docker-entrypoint.sh", "wb") as f:
            f.write(ENTRYPOINT_FLASK.format(service_script=service_script).encode("utf-8"))
        with open(path.parent / "Dockerfile", "w") as f:
            f.write(
                DOCKERFILE.format(
                    base_image=base_image,
                    required_files=" ".join(required_files),
                    required_folders=required_folders,
                    commands="\n".join(commands),
                    env=ENV_FLASK.format(
                        workers=workers,
                        timeout=timeout,
                        worker_class=worker_class,
                        log_level=log_level,
                    ),
                    caching=f"""RUN ["venv/bin/python", "-c", "from {service_script} import app"]
ENV TRANSFORMERS_OFFLINE=1
""",
                )
            )

    @classmethod
    def docker_build_image(cls, tag: str, pull: bool = True, path: str = None, **kwargs):
        """
        Class method to do `docker build ...` in python.
        """
        if path == None:
            path = Path(inspect.getsourcefile(cls))
        else:
            path = Path(path)
        client = docker.from_env()
        image, _ = client.images.build(path=str(path.parent), tag=tag, pull=pull, **kwargs)
        return image

    @classmethod
    def docker_push_image(
        cls,
        repository: str,
        tag: str,
        username: str = None,
        password: str = None,
        **kwargs,
    ):
        """
        Class method to do `docker push ...` in python.
        """
        client = docker.from_env()
        if username is not None and password is not None:
            auth_config = {"username": username, "password": password}
            client.images.push(
                repository=repository,
                tag=tag,
                auth_config=auth_config,
                stream=True,
                **kwargs,
            )
        client.images.push(repository=repository, tag=tag, stream=True, **kwargs)
        return

    @classmethod
    def docker_build_push_image(
        cls,
        repository: str,
        tag: str,
        pull: bool = True,
        username: str = None,
        password: str = None,
        build_kwargs: Dict = {},
        push_kwargs: Dict = {},
    ):
        cls.docker_build_image(tag=f"{repository}:{tag}", pull=pull, **build_kwargs)
        cls.docker_push_image(
            repository=repository,
            tag=tag,
            username=username,
            password=password,
            **push_kwargs,
        )
        return None



from .. import Failure
from .AnnotationsResponse import AnnotationsResponse
from .AudioResponse import AudioResponse
from .ClassificationResponse import ClassificationResponse
from .TextsResponse import TextsResponse


def get_response(response):
    try:
        if "response" in response:
            response = response["response"]
            if response["type"] == "annotations":
                return AnnotationsResponse(**response)
            elif response["type"] == "audio":
                return AudioResponse(**response)
            elif response["type"] == "classification":
                return ClassificationResponse(**response)
            elif response["type"] == "texts":
                return TextsResponse(**response)
            else:
                raise ValueError(f"Response type {response['type']} not known.")
        elif "failure" in response:
            response = response["failure"]
            return Failure(**response)
    except Exception as e:
        if e == ValueError:
            raise e
        print("Response: ", response)
        raise e
